"""CLI commands to post-process a demultiplex run."""

import logging
from pathlib import Path

import click

from cg.cli.utils import CLICK_CONTEXT_SETTINGS
from cg.constants.cli_options import DRY_RUN, FORCE
from cg.models.cg_config import CGConfig
from cg.services.illumina.post_processing.post_processing_service import (
    IlluminaPostProcessingService,
)

LOG = logging.getLogger(__name__)


@click.group(name="finish", context_settings=CLICK_CONTEXT_SETTINGS)
def finish_group():
    """Finish up after demultiplexing."""


@finish_group.command(name="illumina-run")
@click.argument("demultiplexed-run-dir-name")
@DRY_RUN
@click.pass_obj
def post_process_illumina_run(context: CGConfig, demultiplexed_run_dir_name: str, dry_run: bool):
    """Command to post-process an Illumina run after demultiplexing.

    demultiplexed-run-dir-name is the full run name, e.g. '230912_A00187_1009_AHK33MDRX3'.
    """
    post_processing_service = IlluminaPostProcessingService(
        status_db=context.status_db,
        housekeeper_api=context.housekeeper_api,
        dry_run=dry_run,
        demultiplexed_runs_dir=Path(context.run_instruments.illumina.demultiplexed_runs_dir),
    )
    post_processing_service.post_process_illumina_flow_cell(
        sequencing_run_name=demultiplexed_run_dir_name,
    )


@finish_group.command(name="all")
@click.pass_obj
@DRY_RUN
def post_process_all_illumina_runs(context: CGConfig, dry_run: bool):
    """Command to post-process all demultiplexed Illumina runs."""

    post_processing_service = IlluminaPostProcessingService(
        status_db=context.status_db,
        housekeeper_api=context.housekeeper_api,
        dry_run=dry_run,
        demultiplexed_runs_dir=Path(context.run_instruments.illumina.demultiplexed_runs_dir),
    )
    is_error_raised: bool = post_processing_service.post_process_all_runs()
    if is_error_raised:
        raise click.Abort



"""Handler to read data objects."""

import datetime as dt
import logging
from datetime import datetime
from typing import Callable, Iterator, Literal

from sqlalchemy.orm import Query, Session

from cg.constants import SequencingRunDataAvailability, Workflow
from cg.constants.constants import CaseActions, CustomerId, PrepCategory, SampleType
from cg.exc import CaseNotFoundError, CgError, OrderNotFoundError, SampleNotFoundError
from cg.server.dto.orders.orders_request import OrdersRequest
from cg.server.dto.samples.collaborator_samples_request import CollaboratorSamplesRequest
from cg.store.base import BaseHandler
from cg.store.filters.status_analysis_filters import AnalysisFilter, apply_analysis_filter
from cg.store.filters.status_application_filters import ApplicationFilter, apply_application_filter
from cg.store.filters.status_application_limitations_filters import (
    ApplicationLimitationsFilter,
    apply_application_limitations_filter,
)
from cg.store.filters.status_application_version_filters import (
    ApplicationVersionFilter,
    apply_application_versions_filter,
)
from cg.store.filters.status_bed_filters import BedFilter, apply_bed_filter
from cg.store.filters.status_bed_version_filters import BedVersionFilter, apply_bed_version_filter
from cg.store.filters.status_case_filters import CaseFilter, apply_case_filter
from cg.store.filters.status_case_sample_filters import CaseSampleFilter, apply_case_sample_filter
from cg.store.filters.status_collaboration_filters import (
    CollaborationFilter,
    apply_collaboration_filter,
)
from cg.store.filters.status_customer_filters import CustomerFilter, apply_customer_filter
from cg.store.filters.status_illumina_flow_cell_filters import (
    IlluminaFlowCellFilter,
    apply_illumina_flow_cell_filters,
)
from cg.store.filters.status_illumina_metrics_filters import (
    IlluminaMetricsFilter,
    apply_illumina_metrics_filter,
)
from cg.store.filters.status_illumina_sequencing_run_filters import (
    IlluminaSequencingRunFilter,
    apply_illumina_sequencing_run_filter,
)
from cg.store.filters.status_invoice_filters import InvoiceFilter, apply_invoice_filter
from cg.store.filters.status_order_filters import OrderFilter, apply_order_filters
from cg.store.filters.status_organism_filters import OrganismFilter, apply_organism_filter
from cg.store.filters.status_panel_filters import PanelFilter, apply_panel_filter
from cg.store.filters.status_pool_filters import PoolFilter, apply_pool_filter
from cg.store.filters.status_sample_filters import SampleFilter, apply_sample_filter
from cg.store.filters.status_user_filters import UserFilter, apply_user_filter
from cg.store.models import (
    Analysis,
    Application,
    ApplicationLimitations,
    ApplicationVersion,
    Bed,
    BedVersion,
    Case,
    CaseSample,
    Collaboration,
    Customer,
    IlluminaFlowCell,
    IlluminaSampleSequencingMetrics,
    IlluminaSequencingRun,
    Invoice,
    Order,
    Organism,
    Panel,
    Pool,
    Sample,
    SampleRunMetrics,
    User,
)

LOG = logging.getLogger(__name__)


class ReadHandler(BaseHandler):
    """Class for reading items in the database."""

    def __init__(self, session: Session):
        super().__init__(session=session)

    def get_case_by_entry_id(self, entry_id: str) -> Case:
        """Return a case by entry id."""
        cases_query: Query = self._get_query(table=Case)
        return apply_case_filter(
            cases=cases_query, filter_functions=[CaseFilter.BY_ENTRY_ID], entry_id=entry_id
        ).first()

    def has_active_cases_for_sample(self, internal_id: str) -> bool:
        """Check if there are any active cases for a sample."""
        sample = self.get_sample_by_internal_id(internal_id=internal_id)
        active_actions = ["analyze", "running"]

        for family_sample in sample.links:
            case: Case = self.get_case_by_entry_id(entry_id=family_sample.case_id)
            if case.action in active_actions:
                return True

        return False

    def get_application_by_case(self, case_id: str) -> Application:
        """Return the application of a case."""

        return (
            self.get_case_by_internal_id(internal_id=case_id)
            .links[0]
            .sample.application_version.application
        )

    def get_application_limitations_by_tag(self, tag: str) -> list[ApplicationLimitations]:
        """Return application limitations given the application tag."""
        return apply_application_limitations_filter(
            filter_functions=[ApplicationLimitationsFilter.BY_TAG],
            application_limitations=self._get_join_application_limitations_query(),
            tag=tag,
        ).all()

    def get_application_limitation_by_tag_and_workflow(
        self, tag: str, workflow: Workflow
    ) -> ApplicationLimitations | None:
        """Return an application limitation given the application tag and workflow."""
        filter_functions: list[ApplicationLimitationsFilter] = [
            ApplicationLimitationsFilter.BY_TAG,
            ApplicationLimitationsFilter.BY_WORKFLOW,
        ]
        return apply_application_limitations_filter(
            filter_functions=filter_functions,
            application_limitations=self._get_join_application_limitations_query(),
            tag=tag,
            workflow=workflow,
        ).first()

    def get_latest_analysis_to_upload_for_workflow(self, workflow: str = None) -> list[Analysis]:
        """Return latest not uploaded analysis for each case given a workflow."""
        filter_functions: list[AnalysisFilter] = [
            AnalysisFilter.WITH_WORKFLOW,
            AnalysisFilter.IS_NOT_UPLOADED,
        ]
        return apply_analysis_filter(
            filter_functions=filter_functions,
            analyses=self._get_latest_analyses_for_cases_query(),
            workflow=workflow,
        ).all()

    def get_analysis_by_case_entry_id_and_started_at(
        self, case_entry_id: int, started_at_date: dt.datetime
    ) -> Analysis:
        """Fetch an analysis."""
        filter_functions: list[AnalysisFilter] = [
            AnalysisFilter.BY_CASE_ENTRY_ID,
            AnalysisFilter.BY_STARTED_AT,
        ]

        return apply_analysis_filter(
            filter_functions=filter_functions,
            analyses=self._get_query(Analysis),
            case_entry_id=case_entry_id,
            started_at_date=started_at_date,
        ).first()

    def get_cases_by_customer_and_case_name_search(
        self, customer: Customer, case_name_search: str
    ) -> list[Case]:
        """
        Retrieve a list of cases filtered by a customer and matching names.

        Args:
            customer (Customer): The customer object to filter cases by.
            case_name_search (str): The case name search string to filter cases by.

        Returns:
            list[Case]: A list of filtered cases sorted by creation time.
        """
        filter_functions: list[Callable] = [
            CaseFilter.BY_CUSTOMER_ENTRY_ID,
            CaseFilter.BY_NAME_SEARCH,
            CaseFilter.ORDER_BY_CREATED_AT,
        ]

        return apply_case_filter(
            cases=self._get_query(table=Case),
            filter_functions=filter_functions,
            customer_entry_id=customer.id,
            name_search=case_name_search,
        ).all()

    def get_cases_by_customers_action_and_case_search(
        self,
        customers: list[Customer] | None,
        action: str | None,
        case_search: str | None,
        limit: int | None = 30,
    ) -> list[Case]:
        """
        Retrieve a list of cases filtered by customers, action, and matching names or internal ids.

        Args:
            customers (list[Customer] | None): A list of customer objects to filter cases by.
            action (str | None): The action string to filter cases by.
            case_search (str | None): The case search string to filter cases by.
            limit (int | None, default=30): The maximum number of cases to return.

        Returns:
            list[Case]: A list of filtered cases sorted by creation time and limited by the specified number.
        """
        filter_functions: list[Callable] = [
            CaseFilter.BY_CUSTOMER_ENTRY_IDS,
            CaseFilter.BY_ACTION,
            CaseFilter.BY_CASE_SEARCH,
            CaseFilter.ORDER_BY_CREATED_AT,
        ]

        customer_entry_ids: list[int] = (
            [customer.id for customer in customers] if customers else None
        )

        filtered_cases: Query = apply_case_filter(
            cases=self._get_query(table=Case),
            filter_functions=filter_functions,
            action=action,
            case_search=case_search,
            customer_entry_ids=customer_entry_ids,
        )
        return filtered_cases.limit(limit=limit).all()

    def get_cases_by_customer_workflow_and_case_search(
        self,
        customer: Customer | None,
        workflow: str | None,
        case_search: str | None,
        limit: int | None = 30,
    ) -> list[Case]:
        """
        Retrieve a list of cases filtered by customer, workflow, and matching names or internal ids.
        """
        filter_functions: list[Callable] = [
            CaseFilter.BY_CUSTOMER_ENTRY_ID,
            CaseFilter.BY_CASE_SEARCH,
            CaseFilter.WITH_WORKFLOW,
            CaseFilter.ORDER_BY_CREATED_AT,
        ]

        customer_entry_id: int = customer.id if customer else None

        filtered_cases: Query = apply_case_filter(
            cases=self._get_query(table=Case),
            filter_functions=filter_functions,
            case_search=case_search,
            customer_entry_id=customer_entry_id,
            workflow=workflow,
        )
        return filtered_cases.limit(limit=limit).all()

    def get_cases(self) -> list[Case]:
        """Return all cases."""
        return self._get_query(table=Case).all()

    def get_case_samples_by_case_id(self, case_internal_id: str) -> list[CaseSample]:
        """Return the case-sample links associated with a case."""
        return apply_case_sample_filter(
            filter_functions=[CaseSampleFilter.SAMPLES_IN_CASE_BY_INTERNAL_ID],
            case_internal_id=case_internal_id,
            case_samples=self._get_join_case_sample_query(),
        ).all()

    def filter_cases_with_samples(self, case_ids: list[str]) -> list[str]:
        """Return case id:s associated with samples."""
        cases_with_samples = set()
        for case_id in case_ids:
            case: Case = self.get_case_by_internal_id(internal_id=case_id)
            if case and case.links:
                cases_with_samples.add(case_id)
        return list(cases_with_samples)

    def get_cases_by_ticket_id(self, ticket_id: str) -> list[Case]:
        """Return cases associated with a given ticket id."""
        return apply_case_filter(
            cases=self._get_query(table=Case),
            filter_functions=[CaseFilter.BY_TICKET],
            ticket_id=ticket_id,
        ).all()

    def get_customer_id_from_ticket(self, ticket: str) -> str:
        """Returns the customer related to given ticket."""
        cases: list[Case] = self.get_cases_by_ticket_id(ticket_id=ticket)
        if not cases:
            raise ValueError(f"No case found for ticket {ticket}")
        return cases[0].customer.internal_id

    def get_samples_from_ticket(self, ticket: str) -> list[Sample]:
        """Returns the samples related to given ticket."""
        return apply_case_filter(
            cases=self._get_join_sample_family_query(),
            filter_functions=[CaseFilter.BY_TICKET],
            ticket_id=ticket,
        ).all()

    def get_latest_ticket_from_case(self, case_id: str) -> str:
        """Returns the ticket from the most recent sample in a case."""
        return self.get_case_by_internal_id(internal_id=case_id).latest_ticket

    def _is_case_found(self, case: Case, case_id: str) -> None:
        """Raise error if case is false."""
        if not case:
            LOG.error(f"Could not find case {case_id}")
            raise CaseNotFoundError("")

    def get_samples_by_case_id(self, case_id: str) -> list[Sample]:
        """Get samples on a given case id."""

        case: Case = self.get_case_by_internal_id(internal_id=case_id)
        self._is_case_found(case=case, case_id=case_id)
        return case.samples if case else []

    def get_sample_ids_by_case_id(self, case_id: str = None) -> Iterator[str]:
        """Return sample ids from case id."""
        case: Case = self.get_case_by_internal_id(internal_id=case_id)
        self._is_case_found(case=case, case_id=case_id)
        for link in case.links:
            yield link.sample.internal_id

    def get_case_by_name_and_customer(self, customer: Customer, case_name: str) -> Case:
        """Find a case by case name within a customer."""
        return apply_case_filter(
            cases=self._get_query(table=Case),
            filter_functions=[CaseFilter.BY_CUSTOMER_ENTRY_ID, CaseFilter.BY_NAME],
            customer_entry_id=customer.id,
            name=case_name,
        ).first()

    def get_case_by_name(self, name: str) -> Case:
        """Get a case by name."""
        return apply_case_filter(
            cases=self._get_query(table=Case),
            filter_functions=[CaseFilter.BY_NAME],
            name=name,
        ).first()

    def get_sample_by_customer_and_name(
        self, customer_entry_id: list[int], sample_name: str
    ) -> Sample:
        """Get samples within a customer."""
        filter_functions = [
            SampleFilter.BY_CUSTOMER_ENTRY_IDS,
            SampleFilter.BY_SAMPLE_NAME,
        ]

        return apply_sample_filter(
            samples=self._get_query(table=Sample),
            filter_functions=filter_functions,
            customer_entry_ids=customer_entry_id,
            name=sample_name,
        ).first()

    def get_illumina_metrics_entry_by_device_sample_and_lane(
        self, device_internal_id: str, sample_internal_id: str, lane: int
    ) -> IlluminaSampleSequencingMetrics:
        """Get metrics entry by sequencing device internal id, sample internal id and lane."""
        filtered_metrics: Query = apply_illumina_metrics_filter(
            metrics=self._get_joined_illumina_sample_tables(),
            filter_functions=[IlluminaMetricsFilter.BY_LANE],
            lane=lane,
        )
        filtered_flow_cells: Query = apply_illumina_flow_cell_filters(
            flow_cells=filtered_metrics,
            filter_functions=[IlluminaFlowCellFilter.BY_INTERNAL_ID],
            internal_id=device_internal_id,
        )
        filtered_samples: Query = apply_sample_filter(
            samples=filtered_flow_cells,
            filter_functions=[SampleFilter.BY_INTERNAL_ID],
            internal_id=sample_internal_id,
        )
        return filtered_samples.first()

    def get_illumina_sequencing_run_by_device_internal_id(
        self, device_internal_id: str
    ) -> IlluminaSequencingRun:
        """Get Illumina sequencing run entry by device internal id."""
        return apply_illumina_sequencing_run_filter(
            runs=self._get_query(table=IlluminaSequencingRun),
            filter_functions=[IlluminaSequencingRunFilter.BY_DEVICE_INTERNAL_ID],
            device_internal_id=device_internal_id,
        ).first()

    def get_latest_illumina_sequencing_run_for_nipt_case(
        self, case_internal_id: str
    ) -> IlluminaSequencingRun:
        """
        Get Illumina sequencing run entry by case internal id.
        NIPT runs all samples under a single case on the same sequencing run.
        """

        case: Case = self.get_case_by_internal_id(case_internal_id)
        if case.data_analysis != Workflow.FLUFFY:
            raise CgError(f"Case {case_internal_id} is not a NIPT case")
        sequencing_runs: list[IlluminaSequencingRun] = self.get_illumina_sequencing_runs_by_case(
            case.internal_id
        )
        return max(sequencing_runs, key=lambda run: run.sequencing_completed_at)

    def get_illumina_sequencing_runs_by_data_availability(
        self, data_availability: list[str]
    ) -> list[IlluminaSequencingRun] | None:
        """Return Illumina sequencing runs with supplied statuses."""
        return apply_illumina_sequencing_run_filter(
            runs=self._get_query(table=IlluminaSequencingRun),
            data_availability=data_availability,
            filter_functions=[IlluminaSequencingRunFilter.WITH_DATA_AVAILABILITY],
        ).all()

    def get_samples_by_illumina_flow_cell(self, flow_cell_id: str) -> list[Sample] | None:
        """Return samples present on an Illumina flow cell."""
        sequencing_run: IlluminaSequencingRun = (
            self.get_illumina_sequencing_run_by_device_internal_id(device_internal_id=flow_cell_id)
        )
        if sequencing_run:
            return sequencing_run.device.samples

    def get_illumina_sequencing_runs_by_case(
        self, case_id: str
    ) -> list[IlluminaSequencingRun] | None:
        """Get all Illumina sequencing runs for a case."""
        case: Case = self.get_case_by_internal_id(case_id)
        samples_on_case: list[Sample] = case.samples
        sample_metrics: list[SampleRunMetrics] = []
        for sample in samples_on_case:
            sample_metrics.extend(sample.sample_run_metrics)
        sequencing_runs: list[IlluminaSequencingRun] = [
            apply_illumina_sequencing_run_filter(
                runs=self._get_query(IlluminaSequencingRun),
                filter_functions=[IlluminaSequencingRunFilter.BY_ENTRY_ID],
                entry_id=sample_metric.instrument_run_id,
            ).first()
            for sample_metric in sample_metrics
        ]
        return sequencing_runs

    def are_all_illumina_runs_on_disk(self, case_id: str) -> bool:
        """Check if Illumina runs are on disk for sample before starting the analysis."""
        sequencing_runs: list[IlluminaSequencingRun] | None = (
            self.get_illumina_sequencing_runs_by_case(case_id)
        )
        if not sequencing_runs:
            LOG.info("No sequencing runs found")
            return False
        return all(
            sequencing_run.data_availability == SequencingRunDataAvailability.ON_DISK
            for sequencing_run in sequencing_runs
        )

    def request_sequencing_runs_for_case(self, case_id) -> None:
        """Set the status of removed Illumina sequencing runs to REQUESTED for the given case."""
        sequencing_runs: list[IlluminaSequencingRun] | None = (
            self.get_illumina_sequencing_runs_by_case(case_id)
        )
        for sequencing_run in sequencing_runs:
            if sequencing_run.data_availability == SequencingRunDataAvailability.REMOVED:
                sequencing_run.data_availability = SequencingRunDataAvailability.REQUESTED
                LOG.info(
                    f"Setting status for {sequencing_run.device.internal_id} to {SequencingRunDataAvailability.REQUESTED}"
                )
        self.session.commit()

    def get_invoices_by_status(self, is_invoiced: bool = None) -> list[Invoice]:
        """Return invoices by invoiced status."""
        invoices: Query = self._get_query(table=Invoice)
        if is_invoiced:
            return apply_invoice_filter(
                invoices=invoices, filter_functions=[InvoiceFilter.BY_INVOICED]
            ).all()
        else:
            return apply_invoice_filter(
                invoices=invoices, filter_functions=[InvoiceFilter.BY_NOT_INVOICED]
            ).all()

    def get_invoice_by_entry_id(self, entry_id: int) -> Invoice:
        """Return an invoice."""
        invoices: Query = self._get_query(table=Invoice)
        return apply_invoice_filter(
            invoices=invoices,
            entry_id=entry_id,
            filter_functions=[InvoiceFilter.BY_INVOICE_ID],
        ).first()

    def get_pools_and_samples_for_invoice_by_invoice_id(
        self, *, invoice_id: int = None
    ) -> list[Pool | Sample]:
        """Return all pools and samples for an invoice."""
        pools: list[Pool] = apply_pool_filter(
            pools=self._get_query(table=Pool),
            invoice_id=invoice_id,
            filter_functions=[PoolFilter.BY_INVOICE_ID],
        ).all()
        samples: list[Sample] = apply_sample_filter(
            samples=self._get_query(table=Sample),
            invoice_id=invoice_id,
            filter_functions=[SampleFilter.BY_INVOICE_ID],
        ).all()
        return pools + samples

    def get_case_sample_link(self, case_internal_id: str, sample_internal_id: str) -> CaseSample:
        """Return a case-sample link between a family and a sample."""
        filter_functions: list[CaseSampleFilter] = [
            CaseSampleFilter.SAMPLES_IN_CASE_BY_INTERNAL_ID,
            CaseSampleFilter.CASES_WITH_SAMPLE_BY_INTERNAL_ID,
        ]
        return apply_case_sample_filter(
            filter_functions=filter_functions,
            case_samples=self._get_join_case_sample_query(),
            case_internal_id=case_internal_id,
            sample_internal_id=sample_internal_id,
        ).first()

    def new_invoice_id(self) -> int:
        """Fetch invoices."""
        query: Query = self._get_query(table=Invoice)
        ids = [inv.id for inv in query]
        return max(ids) + 1 if ids else 0

    def get_pools_by_customer_id(self, *, customers: list[Customer] | None = None) -> list[Pool]:
        """Return all the pools for a customer."""
        customer_ids = [customer.id for customer in customers]
        return apply_pool_filter(
            pools=self._get_query(table=Pool),
            customer_ids=customer_ids,
            filter_functions=[PoolFilter.BY_CUSTOMER_ID],
        ).all()

    def get_pools_by_name_enquiry(self, *, name_enquiry: str = None) -> list[Pool]:
        """Return all the pools with a name fitting the enquiry."""
        return apply_pool_filter(
            pools=self._get_query(table=Pool),
            name_enquiry=name_enquiry,
            filter_functions=[PoolFilter.BY_NAME_ENQUIRY],
        ).all()

    def get_pools(self) -> list[Pool]:
        """Return all the pools."""
        return self._get_query(table=Pool).all()

    def get_pools_by_order_enquiry(self, *, order_enquiry: str = None) -> list[Pool]:
        """Return all the pools with an order fitting the enquiry."""
        return apply_pool_filter(
            pools=self._get_query(table=Pool),
            order_enquiry=order_enquiry,
            filter_functions=[PoolFilter.BY_ORDER_ENQUIRY],
        ).all()

    def get_pool_by_entry_id(self, entry_id: int) -> Pool:
        """Return a pool by entry id."""
        pools = self._get_query(table=Pool)
        return apply_pool_filter(
            pools=pools, entry_id=entry_id, filter_functions=[PoolFilter.BY_ENTRY_ID]
        ).first()

    def get_pools_to_render(
        self, customers: list[Customer] | None = None, enquiry: str = None
    ) -> list[Pool]:
        pools: list[Pool] = (
            self.get_pools_by_customer_id(customers=customers) if customers else self.get_pools()
        )
        if enquiry:
            pools: list[Pool] = list(
                set(
                    self.get_pools_by_name_enquiry(name_enquiry=enquiry)
                    or set(self.get_pools_by_order_enquiry(order_enquiry=enquiry))
                )
            )
        return pools

    def get_ready_made_library_expected_reads(self, case_id: str) -> int:
        """Return the target reads of a ready made library case."""

        application: Application = self.get_application_by_case(case_id=case_id)

        if application.prep_category != PrepCategory.READY_MADE_LIBRARY.value:
            raise ValueError(
                f"{case_id} is not a ready made library, found prep category: "
                f"{application.prep_category}"
            )
        return application.expected_reads

    def get_samples_by_customer_id_and_pattern(
        self, *, customers: list[Customer] | None = None, pattern: str = None
    ) -> list[Sample]:
        """Get samples by customer and sample internal id  or sample name pattern."""
        samples: Query = self._get_query(table=Sample)
        customer_entry_ids: list[int] = []
        filter_functions: list[SampleFilter] = []
        if customers:
            if not isinstance(customers, list):
                customers = list(customers)
            customer_entry_ids = [customer.id for customer in customers]
            filter_functions.append(SampleFilter.BY_CUSTOMER_ENTRY_IDS)
        if pattern:
            filter_functions.extend([SampleFilter.BY_INTERNAL_ID_OR_NAME_SEARCH])
        filter_functions.append(SampleFilter.ORDER_BY_CREATED_AT_DESC)
        return apply_sample_filter(
            samples=samples,
            customer_entry_ids=customer_entry_ids,
            search_pattern=pattern,
            filter_functions=filter_functions,
        ).all()

    def get_collaborator_samples(self, request: CollaboratorSamplesRequest) -> list[Sample]:
        customer: Customer | None = self.get_customer_by_internal_id(request.customer)
        collaborator_ids = [collaborator.id for collaborator in customer.collaborators]

        filters = [
            SampleFilter.BY_CUSTOMER_ENTRY_IDS,
            SampleFilter.BY_INTERNAL_ID_OR_NAME_SEARCH,
            SampleFilter.ORDER_BY_CREATED_AT_DESC,
            SampleFilter.IS_NOT_CANCELLED,
            SampleFilter.LIMIT,
        ]
        return apply_sample_filter(
            samples=self._get_query(table=Sample),
            customer_entry_ids=collaborator_ids,
            search_pattern=request.enquiry,
            filter_functions=filters,
            limit=request.limit,
        ).all()

    def _get_samples_by_customer_and_subject_id_query(
        self, customer_internal_id: str, subject_id: str
    ) -> Query:
        """Return query of samples of customer with given subject id."""
        records: Query = apply_customer_filter(
            customers=self._get_join_sample_and_customer_query(),
            customer_internal_id=customer_internal_id,
            filter_functions=[CustomerFilter.BY_INTERNAL_ID],
        )
        return apply_sample_filter(
            samples=records,
            subject_id=subject_id,
            filter_functions=[SampleFilter.BY_SUBJECT_ID],
        )

    def get_samples_by_customer_and_subject_id(
        self, customer_internal_id: str, subject_id: str
    ) -> list[Sample]:
        """Get samples of customer with given subject id."""
        return self._get_samples_by_customer_and_subject_id_query(
            customer_internal_id=customer_internal_id, subject_id=subject_id
        ).all()

    def get_samples_by_customer_id_list_and_subject_id_and_is_tumour(
        self, customer_ids: list[int], subject_id: str, is_tumour: bool
    ) -> list[Sample]:
        """Return a list of samples matching a list of customers with given subject id and is a tumour or not."""
        samples = self._get_query(table=Sample)
        filter_functions = [
            SampleFilter.BY_CUSTOMER_ENTRY_IDS,
            SampleFilter.BY_SUBJECT_ID,
        ]
        (
            filter_functions.append(SampleFilter.IS_TUMOUR)
            if is_tumour
            else filter_functions.append(SampleFilter.IS_NOT_TUMOUR)
        )
        return apply_sample_filter(
            samples=samples,
            customer_entry_ids=customer_ids,
            subject_id=subject_id,
            filter_functions=filter_functions,
        ).all()

    def get_samples_by_any_id(self, **identifiers: dict) -> Query:
        """Return a sample query filtered by the given names and values of Sample attributes."""
        samples: Query = self._get_query(table=Sample).order_by(Sample.internal_id.desc())
        for identifier_name, identifier_value in identifiers.items():
            samples: Query = apply_sample_filter(
                filter_functions=[SampleFilter.BY_IDENTIFIER_NAME_AND_VALUE],
                samples=samples,
                identifier_name=identifier_name,
                identifier_value=identifier_value,
            )
        return samples

    def get_sample_by_name(self, name: str) -> Sample:
        """Get sample by name."""
        samples = self._get_query(table=Sample)
        return apply_sample_filter(
            samples=samples, filter_functions=[SampleFilter.BY_SAMPLE_NAME], name=name
        ).first()

    def get_samples_by_type(self, case_id: str, sample_type: SampleType) -> list[Sample] | None:
        """Get samples given a tissue type."""
        samples: Query = apply_case_sample_filter(
            filter_functions=[CaseSampleFilter.SAMPLES_IN_CASE_BY_INTERNAL_ID],
            case_samples=self._get_join_sample_family_query(),
            case_internal_id=case_id,
        )
        samples: Query = apply_sample_filter(
            filter_functions=[SampleFilter.WITH_TYPE],
            samples=samples,
            tissue_type=sample_type,
        )
        return samples.all() if samples else None

    def is_case_down_sampled(self, case_id: str) -> bool:
        """Returns True if all samples in a case are down sampled from another sample."""
        case: Case = self.get_case_by_internal_id(internal_id=case_id)
        return all(sample.from_sample is not None for sample in case.samples)

    def is_case_external(self, case_id: str) -> bool:
        """Returns True if all samples in a case have been sequenced externally."""
        case: Case = self.get_case_by_internal_id(internal_id=case_id)
        return all(sample.application_version.application.is_external for sample in case.samples)

    def get_case_by_internal_id(self, internal_id: str) -> Case | None:
        """Get case by internal id."""
        return apply_case_filter(
            cases=self._get_query(table=Case),
            filter_functions=[CaseFilter.BY_INTERNAL_ID],
            internal_id=internal_id,
        ).first()

    def get_cases_by_internal_ids(self, internal_ids: list[str]) -> list[Case]:
        """Get cases by internal ids."""
        return apply_case_filter(
            cases=self._get_query(table=Case),
            filter_functions=[CaseFilter.BY_INTERNAL_IDS],
            internal_ids=internal_ids,
        ).all()

    def verify_case_exists(self, case_internal_id: str) -> None:
        """Passes silently if case exists in Status DB, raises error if no case or case samples."""

        case: Case = self.get_case_by_internal_id(internal_id=case_internal_id)
        if not case:
            LOG.error(f"Case {case_internal_id} could not be found in Status DB!")
            raise CgError
        if not case.links:
            LOG.error(f"Case {case_internal_id} has no samples in in Status DB!")
            raise CgError
        LOG.info(f"Case {case_internal_id} exists in Status DB")

    def get_running_cases_in_workflow(self, workflow: Workflow) -> list[Case]:
        """Return all running cases in a workflow."""
        return apply_case_filter(
            cases=self._get_query(table=Case),
            filter_functions=[CaseFilter.WITH_WORKFLOW, CaseFilter.IS_RUNNING],
            workflow=workflow,
        ).all()

    def get_not_analysed_cases_by_sample_internal_id(
        self,
        sample_internal_id: str,
    ) -> list[Case]:
        """Get not analysed cases by sample internal id."""

        query: Query = self._get_join_case_and_sample_query()

        not_analysed_cases: Query = apply_case_filter(
            cases=query,
            filter_functions=[
                CaseFilter.NOT_ANALYSED,
            ],
        )

        return apply_sample_filter(
            samples=not_analysed_cases,
            filter_functions=[SampleFilter.BY_INTERNAL_ID],
            internal_id=sample_internal_id,
        ).all()

    def case_with_name_exists(self, case_name: str) -> bool:
        """Check if a case exists in StatusDB."""
        return bool(self.get_case_by_name(case_name))

    def sample_with_id_exists(self, sample_id: str) -> bool:
        """Check if a sample exists in StatusDB."""
        return bool(self.get_sample_by_internal_id(sample_id))

    def get_application_by_tag(self, tag: str) -> Application:
        """Return an application by tag."""
        return apply_application_filter(
            applications=self._get_query(table=Application),
            filter_functions=[ApplicationFilter.BY_TAG],
            tag=tag,
        ).first()

    def get_applications_is_not_archived(self) -> list[Application]:
        """Return applications that are not archived."""
        return (
            apply_application_filter(
                applications=self._get_query(table=Application),
                filter_functions=[ApplicationFilter.IS_NOT_ARCHIVED],
            )
            .order_by(Application.prep_category, Application.tag)
            .all()
        )

    def get_applications(self) -> list[Application]:
        """Return all applications."""
        return (
            self._get_query(table=Application)
            .order_by(Application.prep_category, Application.tag)
            .all()
        )

    def get_current_application_version_by_tag(self, tag: str) -> ApplicationVersion | None:
        """Return the current application version for an application tag."""
        application = self.get_application_by_tag(tag=tag)
        if not application:
            return None
        return apply_application_versions_filter(
            filter_functions=[
                ApplicationVersionFilter.BY_APPLICATION_ENTRY_ID,
                ApplicationVersionFilter.BY_VALID_FROM_BEFORE,
                ApplicationVersionFilter.ORDER_BY_VALID_FROM_DESC,
            ],
            application_versions=self._get_query(table=ApplicationVersion),
            application_entry_id=application.id,
            valid_from=dt.datetime.now(),
        ).first()

    def get_bed_version_by_file_name(self, bed_version_file_name: str) -> BedVersion:
        """Return bed version with file name."""
        return apply_bed_version_filter(
            bed_versions=self._get_query(table=BedVersion),
            bed_version_file_name=bed_version_file_name,
            filter_functions=[BedVersionFilter.BY_FILE_NAME],
        ).first()

    def get_bed_version_by_short_name(self, bed_version_short_name: str) -> BedVersion:
        """Return bed version with short name."""
        return apply_bed_version_filter(
            bed_versions=self._get_query(table=BedVersion),
            bed_version_short_name=bed_version_short_name,
            filter_functions=[BedVersionFilter.BY_SHORT_NAME],
        ).first()

    def get_bed_by_entry_id(self, bed_entry_id: int) -> Bed:
        """Get panel bed with bed entry id."""
        return apply_bed_filter(
            beds=self._get_query(table=Bed),
            filter_functions=[BedFilter.BY_ENTRY_ID],
            bed_entry_id=bed_entry_id,
        ).first()

    def get_bed_by_name(self, bed_name: str) -> Bed:
        """Get panel bed with bed name."""
        return apply_bed_filter(
            beds=self._get_query(table=Bed),
            filter_functions=[BedFilter.BY_NAME],
            bed_name=bed_name,
        ).first()

    def get_active_beds(self) -> Query:
        """Get all beds that are not archived."""
        bed_filter_functions: list[BedFilter] = [
            BedFilter.NOT_ARCHIVED,
            BedFilter.ORDER_BY_NAME,
        ]
        return apply_bed_filter(
            beds=self._get_query(table=Bed), filter_functions=bed_filter_functions
        )

    def get_customer_by_internal_id(self, customer_internal_id: str) -> Customer:
        """Return customer with customer id."""
        return apply_customer_filter(
            filter_functions=[CustomerFilter.BY_INTERNAL_ID],
            customers=self._get_query(table=Customer),
            customer_internal_id=customer_internal_id,
        ).first()

    def get_collaboration_by_internal_id(self, internal_id: str) -> Collaboration:
        """Fetch a customer group by internal id from the store."""
        return apply_collaboration_filter(
            collaborations=self._get_query(table=Collaboration),
            filter_functions=[CollaborationFilter.BY_INTERNAL_ID],
            internal_id=internal_id,
        ).first()

    def get_organism_by_internal_id(self, internal_id: str) -> Organism:
        """Find an organism by internal id."""
        return apply_organism_filter(
            organisms=self._get_query(table=Organism),
            filter_functions=[OrganismFilter.BY_INTERNAL_ID],
            internal_id=internal_id,
        ).first()

    def get_all_organisms(self) -> list[Organism]:
        """Return all organisms ordered by organism internal id."""
        return self._get_query(table=Organism).order_by(Organism.internal_id)

    def get_customers(self) -> list[Customer]:
        """Return costumers."""
        return self._get_query(table=Customer).all()

    def get_panel_by_abbreviation(self, abbreviation: str) -> Panel:
        """Return a panel by abbreviation."""
        return apply_panel_filter(
            panels=self._get_query(table=Panel),
            filters=[PanelFilter.BY_ABBREVIATION],
            abbreviation=abbreviation,
        ).first()

    def get_panels(self) -> list[Panel]:
        """Returns all panels."""
        return self._get_query(table=Panel).order_by(Panel.abbrev).all()

    def get_user_by_email(self, email: str) -> User:
        """Return a user by email from the database."""
        return apply_user_filter(
            users=self._get_query(table=User),
            email=email,
            filter_functions=[UserFilter.BY_EMAIL],
        ).first()

    def get_samples_to_receive(self, external: bool = False) -> list[Sample]:
        """Return samples to receive."""
        records: Query = self._get_join_sample_application_version_query()
        sample_filter_functions: list[SampleFilter] = [
            SampleFilter.IS_NOT_RECEIVED,
            SampleFilter.IS_NOT_DOWN_SAMPLED,
        ]
        records: Query = apply_sample_filter(
            samples=records, filter_functions=sample_filter_functions
        )
        if external:
            records: Query = apply_application_filter(
                applications=records, filter_functions=[ApplicationFilter.IS_EXTERNAL]
            )
        else:
            records: Query = apply_application_filter(
                applications=records,
                filter_functions=[ApplicationFilter.IS_NOT_EXTERNAL],
            )
        return records.order_by(Sample.ordered_at).all()

    def get_samples_to_prepare(self) -> list[Sample]:
        """Return samples to prepare."""
        records: Query = self._get_join_sample_application_version_query()
        sample_filter_functions: list[SampleFilter] = [
            SampleFilter.IS_RECEIVED,
            SampleFilter.IS_NOT_PREPARED,
            SampleFilter.IS_NOT_DOWN_SAMPLED,
            SampleFilter.IS_NOT_SEQUENCED,
        ]
        records: Query = apply_sample_filter(
            samples=records, filter_functions=sample_filter_functions
        )
        records: Query = apply_application_filter(
            applications=records, filter_functions=[ApplicationFilter.IS_NOT_EXTERNAL]
        )

        return records.order_by(Sample.received_at).all()

    def get_samples_to_sequence(self) -> list[Sample]:
        """Return samples in sequencing."""
        records: Query = self._get_join_sample_application_version_query()
        sample_filter_functions: list[SampleFilter] = [
            SampleFilter.IS_PREPARED,
            SampleFilter.IS_NOT_SEQUENCED,
            SampleFilter.IS_NOT_DOWN_SAMPLED,
        ]
        records: Query = apply_sample_filter(
            samples=records, filter_functions=sample_filter_functions
        )
        records: Query = apply_application_filter(
            applications=records, filter_functions=[ApplicationFilter.IS_NOT_EXTERNAL]
        )
        return records.order_by(Sample.prepared_at).all()

    def get_families_with_analyses(self) -> Query:
        """Return all cases in the database with an analysis."""
        return self._get_outer_join_cases_with_analyses_query()

    def get_families_with_samples(self) -> Query:
        """Return all cases in the database with samples."""
        return self._get_join_cases_with_samples_query()

    def cases_to_analyse(self, workflow: Workflow = None, limit: int = None) -> list[Case]:
        """Returns a list if cases ready to be analyzed or set to be reanalyzed."""
        case_filter_functions: list[CaseFilter] = [
            CaseFilter.HAS_SEQUENCE,
            CaseFilter.WITH_WORKFLOW,
            CaseFilter.FOR_ANALYSIS,
        ]
        cases = apply_case_filter(
            cases=self.get_families_with_analyses(),
            filter_functions=case_filter_functions,
            workflow=workflow,
        )

        families: list[Query] = list(cases.order_by(Case.ordered_at))
        families = [
            case_obj
            for case_obj in families
            if case_obj.latest_sequenced
            and (
                case_obj.action == CaseActions.ANALYZE
                or not case_obj.latest_analyzed
                or case_obj.latest_analyzed < case_obj.latest_sequenced
            )
        ]

        return families[:limit]

    def set_case_action(
        self, action: Literal[CaseActions.actions()], case_internal_id: str
    ) -> None:
        """Sets the action of provided cases to None or the given action."""
        case: Case = self.get_case_by_internal_id(internal_id=case_internal_id)
        case.action = action
        self.session.commit()

    def get_cases_to_compress(self, date_threshold: datetime) -> list[Case]:
        """Return all cases that are ready to be compressed by SPRING."""
        case_filter_functions: list[CaseFilter] = [
            CaseFilter.HAS_INACTIVE_ANALYSIS,
            CaseFilter.OLD_BY_CREATION_DATE,
            CaseFilter.IS_COMPRESSIBLE,
        ]
        return apply_case_filter(
            cases=self._get_query(table=Case),
            filter_functions=case_filter_functions,
            creation_date=date_threshold,
        ).all()

    def get_sample_by_entry_id(self, entry_id: int) -> Sample:
        """Return a sample by entry id."""
        sample: Sample | None = apply_sample_filter(
            filter_functions=[SampleFilter.BY_ENTRY_ID],
            samples=self._get_query(table=Sample),
            entry_id=entry_id,
        ).first()

        if not sample:
            LOG.error(f"Could not find sample with entry id {entry_id}")
            raise SampleNotFoundError(f"Could not find sample with entry id {entry_id}")
        return sample

    def get_sample_by_internal_id(self, internal_id: str) -> Sample | None:
        """Return a sample by lims id."""
        return apply_sample_filter(
            filter_functions=[SampleFilter.BY_INTERNAL_ID],
            samples=self._get_query(table=Sample),
            internal_id=internal_id,
        ).first()

    def get_samples_by_identifier(self, object_type: str, identifier: str) -> list[Sample]:
        """Return all samples from a flow cell, case or sample id"""
        object_to_filter: dict[str, Callable] = {
            "sample": self.get_sample_by_internal_id,
            "case": self.get_samples_by_case_id,
            "flow_cell": self.get_samples_by_illumina_flow_cell,
        }
        samples: Sample | list[Sample] = object_to_filter[object_type](identifier)
        return samples if isinstance(samples, list) else [samples]

    def get_samples_by_internal_id(self, internal_id: str) -> list[Sample]:
        """Return all samples by lims id."""
        return apply_sample_filter(
            filter_functions=[SampleFilter.BY_INTERNAL_ID],
            samples=self._get_query(table=Sample),
            internal_id=internal_id,
        ).all()

    def get_analyses_to_upload(self, workflow: Workflow = None) -> list[Analysis]:
        """Return analyses that have not been uploaded."""
        analysis_filter_functions: list[AnalysisFilter] = [
            AnalysisFilter.WITH_WORKFLOW,
            AnalysisFilter.COMPLETED,
            AnalysisFilter.IS_NOT_UPLOADED,
            AnalysisFilter.VALID_IN_PRODUCTION,
            AnalysisFilter.ORDER_BY_COMPLETED_AT,
        ]
        return apply_analysis_filter(
            filter_functions=analysis_filter_functions,
            analyses=self._get_join_analysis_case_query(),
            workflow=workflow,
        ).all()

    def get_analyses_to_clean(
        self, before: datetime = datetime.now(), workflow: Workflow = None
    ) -> list[Analysis]:
        """Return analyses that haven't been cleaned."""
        filter_functions: list[AnalysisFilter] = [
            AnalysisFilter.IS_UPLOADED,
            AnalysisFilter.IS_NOT_CLEANED,
            AnalysisFilter.STARTED_AT_BEFORE,
            AnalysisFilter.CASE_ACTION_IS_NONE,
        ]
        if workflow:
            filter_functions.append(AnalysisFilter.WITH_WORKFLOW)
        return apply_analysis_filter(
            filter_functions=filter_functions,
            analyses=self._get_latest_analyses_for_cases_query(),
            workflow=workflow,
            started_at_date=before,
        ).all()

    def get_analyses_for_case_and_workflow_started_at_before(
        self,
        workflow: Workflow,
        started_at_before: datetime,
        case_internal_id: str,
    ) -> list[Analysis]:
        """Return all analyses older than certain date."""
        case = self.get_case_by_internal_id(internal_id=case_internal_id)
        case_entry_id: int = case.id if case else None
        filter_functions: list[AnalysisFilter] = [
            AnalysisFilter.BY_CASE_ENTRY_ID,
            AnalysisFilter.WITH_WORKFLOW,
            AnalysisFilter.STARTED_AT_BEFORE,
        ]
        return apply_analysis_filter(
            filter_functions=filter_functions,
            analyses=self._get_query(table=Analysis),
            workflow=workflow,
            case_entry_id=case_entry_id,
            started_at_date=started_at_before,
        ).all()

    def get_analyses_for_case_started_at_before(
        self,
        case_internal_id: str,
        started_at_before: datetime,
    ) -> list[Analysis]:
        """Return all analyses for a case older than certain date."""
        case = self.get_case_by_internal_id(internal_id=case_internal_id)
        case_entry_id: int = case.id if case else None
        filter_functions: list[AnalysisFilter] = [
            AnalysisFilter.BY_CASE_ENTRY_ID,
            AnalysisFilter.STARTED_AT_BEFORE,
        ]
        return apply_analysis_filter(
            filter_functions=filter_functions,
            analyses=self._get_query(table=Analysis),
            case_entry_id=case_entry_id,
            started_at_date=started_at_before,
        ).all()

    def get_analyses_for_workflow_started_at_before(
        self, workflow: Workflow, started_at_before: datetime
    ) -> list[Analysis]:
        """Return all analyses for a workflow started before a certain date."""
        filter_functions: list[AnalysisFilter] = [
            AnalysisFilter.WITH_WORKFLOW,
            AnalysisFilter.STARTED_AT_BEFORE,
        ]
        return apply_analysis_filter(
            filter_functions=filter_functions,
            analyses=self._get_query(table=Analysis),
            workflow=workflow,
            started_at_date=started_at_before,
        ).all()

    def get_analyses_started_at_before(self, started_at_before: datetime) -> list[Analysis]:
        """Return all analyses for a workflow started before a certain date."""
        return apply_analysis_filter(
            filter_functions=[AnalysisFilter.STARTED_AT_BEFORE],
            analyses=self._get_query(table=Analysis),
            started_at_date=started_at_before,
        ).all()

    def observations_to_upload(self, workflow: Workflow = None) -> Query:
        """Return observations that have not been uploaded."""
        case_filter_functions: list[CaseFilter] = [
            CaseFilter.WITH_LOQUSDB_SUPPORTED_WORKFLOW,
            CaseFilter.WITH_LOQUSDB_SUPPORTED_SEQUENCING_METHOD,
        ]
        records: Query = apply_case_filter(
            cases=self.get_families_with_samples(),
            filter_functions=case_filter_functions,
            workflow=workflow,
        )
        return apply_sample_filter(
            filter_functions=[SampleFilter.WITHOUT_LOQUSDB_ID], samples=records
        )

    def observations_uploaded(self, workflow: Workflow = None) -> Query:
        """Return observations that have been uploaded."""
        records: Query = apply_case_filter(
            cases=self.get_families_with_samples(),
            filter_functions=[CaseFilter.WITH_LOQUSDB_SUPPORTED_WORKFLOW],
            workflow=workflow,
        )
        records: Query = apply_sample_filter(
            filter_functions=[SampleFilter.WITH_LOQUSDB_ID], samples=records
        )
        return records

    def get_analyses(self) -> list[Analysis]:
        return self._get_query(table=Analysis).all()

    def get_analyses_to_deliver_for_pipeline(self, workflow: Workflow = None) -> list[Analysis]:
        """Return analyses that have been uploaded but not delivered."""
        analyses: Query = apply_sample_filter(
            samples=self._get_join_analysis_sample_family_query(),
            filter_functions=[SampleFilter.IS_NOT_DELIVERED],
        )
        filter_functions: list[AnalysisFilter] = [
            AnalysisFilter.IS_NOT_UPLOADED,
            AnalysisFilter.WITH_WORKFLOW,
            AnalysisFilter.ORDER_BY_UPLOADED_AT,
        ]
        return apply_analysis_filter(
            filter_functions=filter_functions, analyses=analyses, workflow=workflow
        ).all()

    def analyses_to_delivery_report(self, workflow: Workflow | None = None) -> Query:
        """Return analyses that need a delivery report to be regenerated."""
        records: Query = apply_case_filter(
            cases=self._get_join_analysis_case_query(),
            filter_functions=[CaseFilter.REPORT_SUPPORTED],
        )
        analysis_filter_functions: list[AnalysisFilter] = [
            AnalysisFilter.REPORT_BY_WORKFLOW,
            AnalysisFilter.WITHOUT_DELIVERY_REPORT,
            AnalysisFilter.VALID_IN_PRODUCTION,
            AnalysisFilter.ORDER_BY_COMPLETED_AT,
        ]
        return apply_analysis_filter(
            filter_functions=analysis_filter_functions, analyses=records, workflow=workflow
        )

    def analyses_to_upload_delivery_reports(self, workflow: Workflow = None) -> Query:
        """Return analyses that need a delivery report to be uploaded."""
        records: Query = apply_case_filter(
            cases=self._get_join_analysis_case_query(),
            filter_functions=[CaseFilter.WITH_SCOUT_DELIVERY],
        )
        analysis_filter_functions: list[AnalysisFilter] = [
            AnalysisFilter.REPORT_BY_WORKFLOW,
            AnalysisFilter.WITH_DELIVERY_REPORT,
            AnalysisFilter.IS_NOT_UPLOADED,
            AnalysisFilter.VALID_IN_PRODUCTION,
            AnalysisFilter.ORDER_BY_COMPLETED_AT,
        ]
        return apply_analysis_filter(
            filter_functions=analysis_filter_functions, analyses=records, workflow=workflow
        )

    def get_samples_to_deliver(self) -> list[Sample]:
        """Return all samples not delivered."""
        records = self._get_query(table=Sample)
        sample_filter_functions: list[SampleFilter] = [
            SampleFilter.IS_SEQUENCED,
            SampleFilter.IS_NOT_DOWN_SAMPLED,
            SampleFilter.IS_NOT_DELIVERED,
        ]

        records: Query = apply_sample_filter(
            filter_functions=sample_filter_functions,
            samples=records,
        )

        return records.all()

    def get_samples_not_invoiced(self) -> list[Sample]:
        """Return all samples that have not been invoiced, excluding those that
        have been down sampled."""
        records = self._get_query(table=Sample)
        sample_filter_functions: list[SampleFilter] = [
            SampleFilter.HAS_NO_INVOICE_ID,
            SampleFilter.IS_NOT_DOWN_SAMPLED,
        ]

        records: Query = apply_sample_filter(
            filter_functions=sample_filter_functions,
            samples=records,
        )
        return records.all()

    def get_samples_not_down_sampled(self) -> list[Sample]:
        """Return all samples that have not been down sampled."""
        return apply_sample_filter(
            filter_functions=[SampleFilter.IS_NOT_DOWN_SAMPLED],
            samples=self._get_query(table=Sample),
        ).all()

    def get_samples_to_invoice_query(self) -> Query:
        """Return all samples that should be invoiced."""
        sample_filter_functions: list[SampleFilter] = [
            SampleFilter.IS_DELIVERED,
            SampleFilter.HAS_NO_INVOICE_ID,
            SampleFilter.DO_INVOICE,
            SampleFilter.IS_NOT_DOWN_SAMPLED,
        ]
        return apply_sample_filter(
            filter_functions=sample_filter_functions,
            samples=self._get_query(table=Sample),
        )

    def get_pools_to_invoice_query(self) -> Query:
        """Return all pools that should be invoiced."""
        pool_filter_functions: list[PoolFilter] = [
            PoolFilter.IS_DELIVERED,
            PoolFilter.WITHOUT_INVOICE_ID,
            PoolFilter.DO_INVOICE,
        ]
        return apply_pool_filter(
            filter_functions=pool_filter_functions,
            pools=self._get_query(table=Pool),
        )

    def get_samples_to_invoice_for_customer(self, customer: Customer = None) -> list[Sample]:
        """Return all samples that should be invoiced for a customer."""
        return apply_sample_filter(
            samples=self.get_samples_to_invoice_query(),
            filter_functions=[SampleFilter.BY_CUSTOMER],
            customer=customer,
        ).all()

    def get_pools_to_invoice_for_customer(self, customer: Customer = None) -> list[Pool]:
        """Return all pools for a customer that should be invoiced."""
        return apply_pool_filter(
            filter_functions=[PoolFilter.BY_CUSTOMER],
            pools=self.get_pools_to_invoice_query(),
            customer=customer,
        ).all()

    def get_customers_to_invoice(self, records: Query) -> list[Customer]:
        customers_to_invoice: list[Customer] = [
            record.customer
            for record in records.all()
            if record.customer.internal_id != CustomerId.CG_INTERNAL_CUSTOMER
        ]
        return list(set(customers_to_invoice))

    def get_pools_to_receive(self) -> list[Pool]:
        """Return all pools that have been not yet been received."""
        return apply_pool_filter(
            filter_functions=[PoolFilter.IS_NOT_RECEIVED], pools=self._get_query(table=Pool)
        ).all()

    def get_all_pools_to_deliver(self) -> list[Pool]:
        """Return all pools that are received but have not yet been delivered."""
        records = self._get_query(table=Pool)
        pool_filter_functions: list[PoolFilter] = [
            PoolFilter.IS_RECEIVED,
            PoolFilter.IS_NOT_DELIVERED,
        ]

        records: Query = apply_pool_filter(
            filter_functions=pool_filter_functions,
            pools=records,
        )
        return records.all()

    def get_orders(self, orders_request: OrdersRequest) -> tuple[list[Order], int]:
        """Filter, sort and paginate orders based on the provided request."""
        orders: Query = apply_order_filters(
            orders=self._get_query(Order),
            filters=[OrderFilter.BY_WORKFLOW, OrderFilter.BY_SEARCH, OrderFilter.BY_DELIVERED],
            workflow=orders_request.workflow,
            search=orders_request.search,
            delivered=orders_request.delivered,
        )
        total_count: int = orders.count()
        orders: list[Order] = self.sort_and_paginate_orders(
            orders=orders, orders_request=orders_request
        )
        return orders, total_count

    def sort_and_paginate_orders(self, orders: Query, orders_request: OrdersRequest) -> list[Order]:
        return apply_order_filters(
            orders=orders,
            filters=[OrderFilter.SORT, OrderFilter.PAGINATE],
            sort_field=orders_request.sort_field,
            sort_order=orders_request.sort_order,
            page=orders_request.page,
            page_size=orders_request.page_size,
        ).all()

    def get_orders_by_ids(self, order_ids: list[int]) -> list[Order]:
        """Return all orders with the provided ids."""
        return apply_order_filters(
            orders=self._get_query(Order),
            filters=[OrderFilter.BY_IDS],
            ids=order_ids,
        ).all()

    def get_order_by_id(self, order_id: int) -> Order:
        """Returns the entry in Order matching the given id."""
        orders: Query = self._get_query(table=Order)
        order_filter_functions: list[Callable] = [OrderFilter.BY_ID]
        orders: Query = apply_order_filters(
            orders=orders, filters=order_filter_functions, id=order_id
        )
        if not (order := orders.first()):
            raise OrderNotFoundError(f"Order with ID {order_id} not found.")
        return order

    def get_order_by_ticket_id(self, ticket_id: int) -> Order | None:
        """Returns the entry in Order matching the given id."""
        orders: Query = self._get_query(table=Order)
        order_filter_functions: list[Callable] = [OrderFilter.BY_TICKET_ID]
        orders: Query = apply_order_filters(
            orders=orders, filters=order_filter_functions, ticket_id=ticket_id
        )
        return orders.first()

    def get_case_not_received_count(self, order_id: int, cases_to_exclude: list[str]) -> int:
        filters: list[CaseSampleFilter] = [
            CaseSampleFilter.BY_ORDER,
            CaseSampleFilter.CASES_WITH_SAMPLES_NOT_RECEIVED,
            CaseSampleFilter.EXCLUDE_CASES,
        ]
        case_samples: Query = self._join_sample_and_case()
        return apply_case_sample_filter(
            case_samples=case_samples,
            filter_functions=filters,
            order_id=order_id,
            cases_to_exclude=cases_to_exclude,
        ).count()

    def get_case_in_preparation_count(self, order_id: int, cases_to_exclude: list[str]) -> int:
        filters: list[CaseFilter] = [
            CaseSampleFilter.BY_ORDER,
            CaseSampleFilter.CASES_WITH_ALL_SAMPLES_RECEIVED,
            CaseSampleFilter.CASES_WITH_SAMPLES_NOT_PREPARED,
            CaseSampleFilter.EXCLUDE_CASES,
        ]
        case_samples: Query = self._join_sample_and_case()
        return apply_case_sample_filter(
            case_samples=case_samples,
            filter_functions=filters,
            order_id=order_id,
            cases_to_exclude=cases_to_exclude,
        ).count()

    def get_case_in_sequencing_count(self, order_id: int, cases_to_exclude: list[str]) -> int:
        filters: list[CaseSampleFilter] = [
            CaseSampleFilter.BY_ORDER,
            CaseSampleFilter.CASES_WITH_ALL_SAMPLES_RECEIVED,
            CaseSampleFilter.CASES_WITH_ALL_SAMPLES_PREPARED,
            CaseSampleFilter.CASES_WITH_SAMPLES_NOT_SEQUENCED,
            CaseSampleFilter.EXCLUDE_CASES,
        ]
        case_samples: Query = self._join_sample_and_case()
        return apply_case_sample_filter(
            case_samples=case_samples,
            filter_functions=filters,
            order_id=order_id,
            cases_to_exclude=cases_to_exclude,
        ).count()

    def get_case_failed_sequencing_count(self, order_id: int, cases_to_exclude: list[str]) -> int:
        filters: list[CaseSampleFilter] = [
            CaseSampleFilter.BY_ORDER,
            CaseSampleFilter.CASES_WITH_ALL_SAMPLES_RECEIVED,
            CaseSampleFilter.CASES_WITH_ALL_SAMPLES_PREPARED,
            CaseSampleFilter.CASES_WITH_ALL_SAMPLES_SEQUENCED,
            CaseSampleFilter.CASES_FAILED_SEQUENCING_QC,
            CaseSampleFilter.EXCLUDE_CASES,
        ]
        case_samples: Query = self._join_sample_and_case()
        return apply_case_sample_filter(
            case_samples=case_samples,
            filter_functions=filters,
            order_id=order_id,
            cases_to_exclude=cases_to_exclude,
        ).count()

    def get_illumina_flow_cell_by_internal_id(self, internal_id: str) -> IlluminaFlowCell:
        """Return a flow cell by internal id."""
        return apply_illumina_flow_cell_filters(
            filter_functions=[IlluminaFlowCellFilter.BY_INTERNAL_ID],
            flow_cells=self._get_query(table=IlluminaFlowCell),
            internal_id=internal_id,
        ).first()

    def get_cases_for_sequencing_qc(self) -> list[Case]:
        """Return all cases that are ready for sequencing QC."""
        query = (
            self._get_query(table=Case)
            .join(Case.links)
            .join(CaseSample.sample)
            .join(ApplicationVersion)
            .join(Application)
        )
        return apply_case_filter(
            cases=query,
            filter_functions=[
                CaseFilter.PENDING_OR_FAILED_SEQUENCING_QC,
                CaseFilter.HAS_SEQUENCE,
            ],
        ).all()

    def get_case_ids_with_sample(self, sample_id: int) -> list[str]:
        """Return all case ids with a sample."""
        sample: Sample = self.get_sample_by_entry_id(sample_id)
        return [link.case.internal_id for link in sample.links] if sample else []

    def get_case_ids_for_samples(self, sample_ids: list[int]) -> list[str]:
        case_ids: list[str] = []
        for sample_id in sample_ids:
            case_ids.extend(self.get_case_ids_with_sample(sample_id))
        return list(set(case_ids))



from pathlib import Path
import logging

from cg.constants.file_transfer_service import PACBIO_MANIFEST_FILE_PATTERN, TRANSFER_VALIDATED_FILE
from cg.constants.constants import FileFormat
from cg.io.controller import WriteFile
from cg.models.cg_config import CGConfig
from cg.services.validate_file_transfer_service.validate_file_transfer_service import (
    ValidateFileTransferService,
)


LOG = logging.getLogger(__name__)


class ValidatePacbioFileTransferService(ValidateFileTransferService):

    def __init__(self, config: CGConfig):
        super().__init__()
        self.config: CGConfig = config
        self.data_dir: Path = Path(self.config.run_instruments.pacbio.data_dir)
        self.trigger_dir: Path = Path(self.config.run_instruments.pacbio.systemd_trigger_dir)

    @staticmethod
    def get_run_id(manifest_file_path: Path) -> str:
        """
        Get the run name from the path of the manifest file.
        Example: r84202_20240307_145215/1_C01/metadata/m84202_240307_145611_s3.transferdone
        will return r84202_20240307_145215.
        """
        return manifest_file_path.parts[-4]

    def get_smrt_cell_id(self, manifest_file_path: Path) -> str:
        """
        Get the run name from the path of the manifest file.
        Example: r84202_20240307_145215/1_C01/metadata/m84202_240307_145611_s3.transferdone
        will return 1_C01.
        """
        return self.get_smrt_cell_path(manifest_file_path).name

    @staticmethod
    def get_smrt_cell_path(manifest_file_path: Path) -> Path:
        """
        Get the run name from the path of the manifest file.
        Example: r84202_20240307_145215/1_C01/metadata/m84202_240307_145611_s3.transferdone
        will return r84202_20240307_145215/1_C01.
        """
        return Path(manifest_file_path.parts[-3])

    @staticmethod
    def transfer_validated_file_name(manifest_file_path: Path) -> str:
        return f"{manifest_file_path.parent}/{TRANSFER_VALIDATED_FILE}"

    def create_validated_transfer_file(self, manifest_file_path: Path) -> None:
        file_name: Path = Path(self.transfer_validated_file_name(manifest_file_path))
        writer = WriteFile()
        writer.write_file_from_content(file_path=file_name, content="", file_format=FileFormat.TXT)
        LOG.debug(f"Created validated transfer file {file_name}")

    def is_transfer_validated(self, manifest_file_path: Path) -> bool:
        return Path(self.transfer_validated_file_name(manifest_file_path)).exists()

    def create_systemd_trigger_file(self, manifest_file_path: Path) -> None:
        systemd_trigger_file_name: Path = Path(
            self.trigger_dir,
            f"{self.get_run_id(manifest_file_path)}-{self.get_smrt_cell_id(manifest_file_path)}",
        )

        writer = WriteFile()
        writer.write_file_from_content(
            file_path=systemd_trigger_file_name, content="", file_format=FileFormat.TXT
        )
        LOG.debug(f"Created systemd trigger file {systemd_trigger_file_name}")

    def validate_transfer_done(self, manifest_file_path: Path) -> None:
        if not self.is_transfer_completed(
            manifest_file=manifest_file_path,
            source_dir=self.get_smrt_cell_path(manifest_file_path),
            manifest_file_format=FileFormat.TXT,
        ):
            LOG.info(
                f"Transfer not done for run {self.get_run_id(manifest_file_path)} and smrt cell {self.get_smrt_cell_id(manifest_file_path)}"
            )
            return
        self.create_validated_transfer_file(manifest_file_path)
        self.create_systemd_trigger_file(manifest_file_path)
        LOG.info(
            f"Transfer validated for run {self.get_run_id(manifest_file_path)} and smrt cell {self.get_smrt_cell_id(manifest_file_path)}"
        )

    def validate_all_transfer_done(self) -> None:
        manifest_file_paths = self.get_manifest_file_paths(
            source_dir=self.data_dir, pattern=PACBIO_MANIFEST_FILE_PATTERN
        )
        for manifest_file_path in manifest_file_paths:
            if not self.is_transfer_validated(manifest_file_path):
                self.validate_transfer_done(manifest_file_path)



"""Tests for the housekeeper storage functions of the demultiplexing post post-processing module."""

from pathlib import Path
from unittest.mock import PropertyMock, patch

from housekeeper.store.models import File

from cg.apps.demultiplex.sample_sheet.sample_sheet_models import SampleSheet
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants.demultiplexing import DemultiplexingDirsAndFiles
from cg.constants.housekeeper_tags import SequencingFileTag
from cg.models.cg_config import CGConfig
from cg.models.run_devices.illumina_run_directory_data import IlluminaRunDirectoryData
from cg.services.illumina.post_processing.housekeeper_storage import (
    add_demux_logs_to_housekeeper,
    add_run_parameters_file_to_housekeeper,
    add_sample_fastq_files_to_housekeeper,
)
from cg.store.store import Store


def test_add_fastq_files_to_housekeeper(
    new_demultiplex_context: CGConfig,
    novaseq_6000_post_1_5_kits_flow_cell: IlluminaRunDirectoryData,
    novaseq_6000_post_1_5_kits_sample_sheet_with_selected_samples: SampleSheet,
    selected_novaseq_6000_post_1_5_kits_sample_ids: list[str],
    tmp_demultiplexed_novaseq_6000_post_1_5_kits_path: Path,
):
    """Test that the fastq files of a sequencing run are added to Housekeeper."""
    # GIVEN a Housekeeper API and a Store
    hk_api: HousekeeperAPI = new_demultiplex_context.housekeeper_api
    store: Store = new_demultiplex_context.status_db

    # GIVEN that there are no fastq files in housekeeper for the selected samples
    for sample_id in selected_novaseq_6000_post_1_5_kits_sample_ids:
        assert hk_api.get_files(tags=[SequencingFileTag.FASTQ], bundle=sample_id).count() == 0

    # GIVEN a demultiplexed run dir data with a sample sheet and samples
    with patch.object(
        target=IlluminaRunDirectoryData,
        attribute="sample_sheet",
        new_callable=PropertyMock(
            return_value=novaseq_6000_post_1_5_kits_sample_sheet_with_selected_samples
        ),
    ), patch.object(
        target=IlluminaRunDirectoryData,
        attribute="get_demultiplexed_runs_dir",
        return_value=tmp_demultiplexed_novaseq_6000_post_1_5_kits_path,
    ):

        # WHEN adding the sample fastq files to Housekeeper
        add_sample_fastq_files_to_housekeeper(
            run_directory_data=novaseq_6000_post_1_5_kits_flow_cell,
            hk_api=hk_api,
            store=store,
        )

    # THEN two fastq files per sample were added to Housekeeper
    for sample_id in selected_novaseq_6000_post_1_5_kits_sample_ids:
        fastq_files: list[File] = hk_api.get_files(
            tags=[SequencingFileTag.FASTQ],
            bundle=sample_id,
        ).all()
        assert fastq_files
        assert len(fastq_files) == 2
        assert isinstance(fastq_files[0], File)


def test_add_demux_logs_to_housekeeper(
    new_demultiplex_context: CGConfig,
    novaseq_6000_post_1_5_kits_flow_cell: IlluminaRunDirectoryData,
    demultiplex_log_file_names: list[str],
):
    """Test that the demultiplex log files of a sequencing run are added to Housekeeper."""
    # GIVEN a Housekeeper API
    hk_api: HousekeeperAPI = new_demultiplex_context.housekeeper_api

    # GIVEN a bundle and flow cell version exists in Housekeeper
    hk_api.add_bundle_and_version_if_non_existent(
        bundle_name=novaseq_6000_post_1_5_kits_flow_cell.id
    )

    # GIVEN that there are no log files in Housekeeper
    existing_logs: list[File] = hk_api.get_files(
        tags=[SequencingFileTag.DEMUX_LOG],
        bundle=novaseq_6000_post_1_5_kits_flow_cell.id,
    ).all()
    assert not existing_logs

    # WHEN adding the demux logs to Housekeeper
    add_demux_logs_to_housekeeper(
        run_directory_data=novaseq_6000_post_1_5_kits_flow_cell,
        hk_api=hk_api,
    )

    # THEN the demux log files were added to Housekeeper
    files: list[File] = hk_api.get_files(
        tags=[SequencingFileTag.DEMUX_LOG],
        bundle=novaseq_6000_post_1_5_kits_flow_cell.id,
    ).all()
    assert len(files) == 2
    for file in files:
        assert file.path.split("/")[-1] in demultiplex_log_file_names


def test_add_run_parameters_to_housekeeper(
    new_demultiplex_context: CGConfig, novaseq_x_flow_cell: IlluminaRunDirectoryData
):
    """Test that the run parameters file of a sequencing run is added to Housekeeper."""
    # GIVEN a flow cell with a run parameters file and a Housekeeper API
    hk_api = new_demultiplex_context.housekeeper_api

    # GIVEN that a run parameters file does not exist for the flow cell in Housekeeper
    assert not hk_api.files(tags=[SequencingFileTag.RUN_PARAMETERS, novaseq_x_flow_cell.id]).all()

    # GIVEN that a bundle and version exists in Housekeeper
    hk_api.add_bundle_and_version_if_non_existent(bundle_name=novaseq_x_flow_cell.id)

    # WHEN adding the run parameters file to Housekeeper
    add_run_parameters_file_to_housekeeper(
        run_directory_data=novaseq_x_flow_cell,
        hk_api=hk_api,
    )

    # THEN the run parameters file was added to Housekeeper
    run_parameters_file: File = hk_api.files(
        tags=[SequencingFileTag.RUN_PARAMETERS, novaseq_x_flow_cell.id]
    ).first()
    assert run_parameters_file.path.endswith(DemultiplexingDirsAndFiles.RUN_PARAMETERS_PASCAL_CASE)



from datetime import datetime

from pydantic import BaseModel

from cg.clients.freshdesk.constants import Priority, Source, Status


class TicketCreate(BaseModel):
    """Freshdesk ticket."""

    attachments: list[dict[str, str]] = []
    email: str
    email_config_id: int | None = None
    name: str
    priority: int = Priority.LOW
    source: int = Source.EMAIL
    status: int = Status.OPEN
    subject: str
    tags: list[str] = []
    type: str | None = None


class TicketResponse(BaseModel):
    """Freshdesk ticket response."""

    attachments: list[dict[str, str]] = []
    created_at: datetime | None = None
    email: str
    id: int
    name: str | None = None
    priority: int
    source: int
    status: int
    subject: str
    tags: list[str] = []
    type: str | None = None



from pathlib import Path

import pytest
from _pytest.fixtures import FixtureRequest
from click.testing import CliRunner, Result
from pydantic import BaseModel

from cg.apps.demultiplex.sample_sheet.api import IlluminaSampleSheetService
from cg.apps.demultiplex.sample_sheet.sample_models import IlluminaSampleIndexSetting
from cg.cli.demultiplex.sample_sheet import create_sheet
from cg.constants.process import EXIT_SUCCESS
from cg.io.txt import read_txt
from cg.models.cg_config import CGConfig
from cg.models.run_devices.illumina_run_directory_data import IlluminaRunDirectoryData

GET_FLOW_CELL_SAMPLES: str = "cg.apps.demultiplex.sample_sheet.api.get_flow_cell_samples"


def test_create_sample_sheet_no_run_parameters_fails(
    cli_runner: CliRunner,
    tmp_flow_cell_without_run_parameters_path: Path,
    sample_sheet_context_broken_flow_cells: CGConfig,
    hiseq_2500_custom_index_bcl_convert_lims_samples: list[IlluminaSampleIndexSetting],
    caplog,
    mocker,
):
    """Test that creating a sequencing run sample sheet fails if there is no run parameters file."""
    # GIVEN a sequencing run directory with a non-existing sample sheet nor RunParameters file
    flow_cell = IlluminaRunDirectoryData(tmp_flow_cell_without_run_parameters_path)

    # GIVEN that the context's sequencing run directory holds the given sequencing run
    assert (
        sample_sheet_context_broken_flow_cells.run_instruments.illumina.demultiplexed_runs_dir
        == flow_cell.path.parent.as_posix()
    )

    # GIVEN sequencing run samples
    mocker.patch(
        GET_FLOW_CELL_SAMPLES,
        return_value=hiseq_2500_custom_index_bcl_convert_lims_samples,
    )

    # WHEN running the create sample sheet command
    result: Result = cli_runner.invoke(
        create_sheet, [flow_cell.full_name], obj=sample_sheet_context_broken_flow_cells
    )

    # THEN the process exits with a non-zero exit code
    assert result.exit_code != EXIT_SUCCESS

    # THEN the correct information is communicated
    assert "No run parameters file found in sequencing run" in caplog.text


class SampleSheetScenario(BaseModel):
    sequencing_run_directory: str
    lims_samples: str
    correct_sample_sheet: str


@pytest.mark.parametrize(
    "scenario",
    [
        SampleSheetScenario(
            sequencing_run_directory="tmp_novaseq_6000_pre_1_5_kits_flow_cell_without_sample_sheet_path",
            lims_samples="novaseq_6000_pre_1_5_kits_bcl_convert_lims_samples",
            correct_sample_sheet="novaseq_6000_pre_1_5_kits_correct_sample_sheet_path",
        ),
        SampleSheetScenario(
            sequencing_run_directory="tmp_novaseq_6000_post_1_5_kits_flow_cell_without_sample_sheet_path",
            lims_samples="novaseq_6000_post_1_5_kits_bcl_convert_lims_samples",
            correct_sample_sheet="novaseq_6000_post_1_5_kits_correct_sample_sheet_path",
        ),
        SampleSheetScenario(
            sequencing_run_directory="tmp_novaseq_x_without_sample_sheet_flow_cell_path",
            lims_samples="novaseq_x_lims_samples",
            correct_sample_sheet="novaseq_x_correct_sample_sheet",
        ),
    ],
    ids=[
        "Old NovaSeq 6000 sequencing run",
        "New NovaSeq 6000 sequencing run",
        "NovaSeq X sequencing run",
    ],
)
def test_create_v2_sample_sheet(
    cli_runner: CliRunner,
    scenario: SampleSheetScenario,
    sample_sheet_context: CGConfig,
    mocker,
    request: FixtureRequest,
):
    """Test that creating a v2 sample sheet works."""
    # GIVEN a sample sheet context with a sample sheet api
    sample_sheet_api: IlluminaSampleSheetService = sample_sheet_context.sample_sheet_api

    # GIVEN a sequencing run directory with some run parameters
    flow_cell_directory: Path = request.getfixturevalue(scenario.sequencing_run_directory)
    flow_cell = IlluminaRunDirectoryData(flow_cell_directory)
    assert flow_cell.run_parameters_path.exists()

    # GIVEN that there is no sample sheet in the sequencing run dir
    assert not flow_cell.sample_sheet_exists()

    # GIVEN that there are no sample sheet in Housekeeper
    assert not sample_sheet_context.housekeeper_api.get_sample_sheets_from_latest_version(
        flow_cell.id
    )

    # GIVEN flow cell samples
    mocker.patch(
        GET_FLOW_CELL_SAMPLES,
        return_value=request.getfixturevalue(scenario.lims_samples),
    )
    # GIVEN a LIMS API that returns samples

    # WHEN creating a sample sheet
    result: Result = cli_runner.invoke(
        create_sheet,
        [str(flow_cell_directory)],
        obj=sample_sheet_context,
    )

    # THEN the process finishes successfully
    assert result.exit_code == EXIT_SUCCESS

    # THEN the sample sheet was created
    assert flow_cell.sample_sheet_exists()

    # THEN the sample sheet passes validation
    sample_sheet_api.validate_sample_sheet(flow_cell.sample_sheet_path)

    # THEN the sample sheet is in Housekeeper
    assert sample_sheet_context.housekeeper_api.get_sample_sheets_from_latest_version(flow_cell.id)

    # THEN the sample sheet should have the exact structure
    generated_content: str = read_txt(flow_cell.sample_sheet_path)
    correct_content: str = read_txt(Path(request.getfixturevalue(scenario.correct_sample_sheet)))
    assert generated_content == correct_content



from datetime import datetime

from sqlalchemy.orm import Query

from cg.constants.constants import CaseActions, DataDelivery, Workflow
from cg.constants.sequencing import SequencingMethod
from cg.constants.subject import PhenotypeStatus
from cg.store.filters.status_case_filters import (
    filter_case_by_internal_id,
    filter_cases_by_case_search,
    filter_cases_by_customer_entry_ids,
    filter_cases_by_entry_id,
    filter_cases_by_name,
    filter_cases_by_priority,
    filter_cases_by_ticket_id,
    filter_cases_by_workflow_search,
    filter_cases_for_analysis,
    filter_cases_has_sequence,
    filter_cases_not_analysed,
    filter_cases_with_loqusdb_supported_sequencing_method,
    filter_cases_with_loqusdb_supported_workflow,
    filter_cases_with_scout_data_delivery,
    filter_cases_with_workflow,
    filter_inactive_analysis_cases,
    filter_newer_cases_by_order_date,
    filter_older_cases_by_creation_date,
    filter_report_supported_data_delivery_cases,
    filter_running_cases,
)
from cg.store.models import Analysis, Case, CaseSample, Sample
from cg.store.store import Store
from tests.store_helpers import StoreHelpers


def test_filter_cases_has_sequence(
    base_store: Store, helpers: StoreHelpers, timestamp_now: datetime
):
    """Test that a case is returned when there is a cases with a sequenced sample."""

    # GIVEN a sequenced sample
    test_sample: Sample = helpers.add_sample(base_store, last_sequenced_at=timestamp_now)

    # GIVEN a case
    test_case = helpers.add_case(base_store)

    # GIVEN a database with a case with one sequenced samples for specified analysis
    link = base_store.relate_sample(test_case, test_sample, PhenotypeStatus.UNKNOWN)
    base_store.session.add(link)

    # GIVEN a cases Query
    cases: Query = base_store._get_outer_join_cases_with_analyses_query()

    # WHEN getting cases to analyse
    cases: Query = filter_cases_has_sequence(cases=cases)

    # ASSERT that cases is a query
    assert isinstance(cases, Query)

    # THEN cases should contain the test case
    assert cases


def test_filter_cases_has_sequence_when_external(base_store: Store, helpers: StoreHelpers):
    """Test that a case is returned when there is a case with an externally sequenced sample."""

    # GIVEN a sequenced sample
    test_sample: Sample = helpers.add_sample(base_store, is_external=True, last_sequenced_at=None)

    # GIVEN a case
    test_case = helpers.add_case(base_store)

    # GIVEN a database with a case with one sequenced samples for specified analysis
    link = base_store.relate_sample(test_case, test_sample, PhenotypeStatus.UNKNOWN)
    base_store.session.add(link)

    # GIVEN a cases Query
    cases: Query = base_store._get_outer_join_cases_with_analyses_query()

    # WHEN getting cases to analyse
    cases: Query = filter_cases_has_sequence(cases=cases)

    # ASSERT that cases is a query
    assert isinstance(cases, Query)

    # THEN cases should contain the test case
    assert cases


def test_filter_cases_has_sequence_when_not_sequenced(base_store: Store, helpers: StoreHelpers):
    """Test that no case is returned when there is a cases with sample that has not been sequenced."""

    # GIVEN a sequenced sample
    test_sample: Sample = helpers.add_sample(base_store, last_sequenced_at=None)

    # GIVEN a case
    test_case = helpers.add_case(base_store)

    # GIVEN a database with a case with one sequenced samples for specified analysis
    link = base_store.relate_sample(test_case, test_sample, PhenotypeStatus.UNKNOWN)
    base_store.session.add(link)

    # GIVEN a cases Query
    cases: Query = base_store._get_outer_join_cases_with_analyses_query()

    # WHEN getting cases to analyse
    cases: Query = filter_cases_has_sequence(cases=cases)

    # ASSERT that cases is a query
    assert isinstance(cases, Query)

    # THEN cases should not contain the test case
    assert not cases.all()


def test_filter_cases_has_sequence_when_not_external_nor_sequenced(
    base_store: Store, helpers: StoreHelpers
):
    """Test that no case is returned when there is a cases with sample that has not been sequenced nor is external."""

    # GIVEN a sequenced sample
    test_sample: Sample = helpers.add_sample(base_store, is_external=False, last_sequenced_at=None)

    # GIVEN a case
    test_case = helpers.add_case(base_store)

    # GIVEN a database with a case with one sequenced samples for specified analysis
    link = base_store.relate_sample(test_case, test_sample, PhenotypeStatus.UNKNOWN)
    base_store.session.add(link)

    # GIVEN a cases Query
    cases: Query = base_store._get_outer_join_cases_with_analyses_query()

    # WHEN getting cases to analyse
    cases: Query = filter_cases_has_sequence(cases=cases)

    # ASSERT that cases is a query
    assert isinstance(cases, Query)

    # THEN cases should not contain the test case
    assert not cases.all()


def test_filter_cases_with_workflow_when_correct_workflow(
    base_store: Store, helpers: StoreHelpers, timestamp_now: datetime
):
    """Test that no case is returned when there are no cases with the specified workflow."""

    # GIVEN a sequenced sample
    test_sample: Sample = helpers.add_sample(base_store, last_sequenced_at=timestamp_now)

    # GIVEN a cancer case
    test_case = helpers.add_case(base_store, data_analysis=Workflow.BALSAMIC)

    # GIVEN a database with a case with one sequenced samples for specified analysis
    link = base_store.relate_sample(test_case, test_sample, PhenotypeStatus.UNKNOWN)
    base_store.session.add(link)

    # GIVEN a cases Query
    cases: Query = base_store._get_outer_join_cases_with_analyses_query()

    # WHEN getting cases to analyse for another workflow
    cases: list[Query] = list(filter_cases_with_workflow(cases=cases, workflow=Workflow.BALSAMIC))

    # THEN cases should contain the test case
    assert cases


def test_filter_cases_with_workflow_when_incorrect_pipline(
    base_store: Store, helpers: StoreHelpers, timestamp_now: datetime
):
    """Test that no case is returned when there are no cases with the specified workflow."""

    # GIVEN a sequenced sample
    test_sample: Sample = helpers.add_sample(base_store, last_sequenced_at=timestamp_now)

    # GIVEN a cancer case
    test_case: Case = helpers.add_case(base_store, data_analysis=Workflow.BALSAMIC)

    # GIVEN a database with a case with one sequenced samples for specified analysis
    link = base_store.relate_sample(test_case, test_sample, PhenotypeStatus.UNKNOWN)
    base_store.session.add(link)

    # GIVEN a cases Query
    cases: Query = base_store._get_outer_join_cases_with_analyses_query()

    # WHEN getting cases to analyse for another workflow
    cases: list[Query] = list(filter_cases_with_workflow(cases=cases, workflow=Workflow.MIP_DNA))

    # THEN cases should not contain the test case
    assert not cases


def test_filter_cases_with_loqusdb_supported_workflow(
    base_store: Store, helpers: StoreHelpers, timestamp_now: datetime
):
    """Test retrieval of cases that support Loqusdb upload."""

    # GIVEN a sequenced sample
    test_sample: Sample = helpers.add_sample(base_store, last_sequenced_at=timestamp_now)

    # GIVEN a MIP-DNA and a FLUFFY case
    test_mip_case: Case = helpers.add_case(base_store, data_analysis=Workflow.MIP_DNA)
    test_mip_case.customer.loqus_upload = True
    test_fluffy_case: Case = helpers.add_case(
        base_store, name="test", data_analysis=Workflow.FLUFFY
    )
    test_fluffy_case.customer.loqus_upload = True

    # GIVEN a database with a case with one sequenced samples for specified analysis
    link_1: CaseSample = base_store.relate_sample(
        test_mip_case, test_sample, PhenotypeStatus.UNKNOWN
    )
    link_2: CaseSample = base_store.relate_sample(
        test_fluffy_case, test_sample, PhenotypeStatus.UNKNOWN
    )
    base_store.session.add_all([link_1, link_2])

    # GIVEN a cases Query
    cases: Query = base_store._get_outer_join_cases_with_analyses_query()

    # WHEN getting cases with workflow
    cases: list[Query] = list(filter_cases_with_loqusdb_supported_workflow(cases=cases))

    # THEN only the Loqusdb supported case should be extracted
    assert test_mip_case in cases
    assert test_fluffy_case not in cases


def test_filter_cases_with_loqusdb_supported_sequencing_method(
    base_store: Store, helpers: StoreHelpers, timestamp_now: datetime
):
    """Test retrieval of cases with a valid Loqusdb sequencing method."""

    # GIVEN a sample with a valid Loqusdb sequencing method
    test_sample_wes: Sample = helpers.add_sample(
        base_store, application_type=SequencingMethod.WES, last_sequenced_at=timestamp_now
    )

    # GIVEN a MIP-DNA associated test case
    test_case_wes: Case = helpers.add_case(base_store, data_analysis=Workflow.MIP_DNA)
    link = base_store.relate_sample(test_case_wes, test_sample_wes, PhenotypeStatus.UNKNOWN)
    base_store.session.add(link)

    # GIVEN a cases Query
    cases: Query = base_store._get_outer_join_cases_with_analyses_query()

    # WHEN retrieving the available cases
    cases: Query = filter_cases_with_loqusdb_supported_sequencing_method(
        cases=cases, workflow=Workflow.MIP_DNA
    )

    # ASSERT that cases is a query
    assert isinstance(cases, Query)

    # THEN the expected case should be retrieved
    assert test_case_wes in cases


def test_filter_cases_with_loqusdb_supported_sequencing_method_empty(
    base_store: Store, helpers: StoreHelpers, timestamp_now: datetime
):
    """Test retrieval of cases with a valid Loqusdb sequencing method."""

    # GIVEN a not supported loqusdb sample
    test_sample_wts: Sample = helpers.add_sample(
        base_store, is_rna=True, name="sample_wts", last_sequenced_at=timestamp_now
    )

    # GIVEN a MIP-DNA associated test case
    test_case_wts: Case = helpers.add_case(base_store, data_analysis=Workflow.MIP_DNA)
    link = base_store.relate_sample(test_case_wts, test_sample_wts, PhenotypeStatus.UNKNOWN)
    base_store.session.add(link)

    # GIVEN a cases Query
    cases: Query = base_store._get_outer_join_cases_with_analyses_query()

    # WHEN retrieving the valid cases
    cases: Query = filter_cases_with_loqusdb_supported_sequencing_method(
        cases=cases, workflow=Workflow.MIP_DNA
    )

    # ASSERT that cases is a query
    assert isinstance(cases, Query)

    # THEN no cases should be returned
    assert not cases.all()


def test_filter_cases_for_analysis(
    base_store: Store, helpers: StoreHelpers, timestamp_now: datetime
):
    """Test that a case is returned when there is a cases with an action set to analyse."""

    # GIVEN a sequenced sample
    test_sample: Sample = helpers.add_sample(base_store, last_sequenced_at=timestamp_now)

    # GIVEN a completed analysis
    test_analysis: Analysis = helpers.add_analysis(
        base_store, completed_at=timestamp_now, workflow=Workflow.MIP_DNA
    )

    # Given an action set to analyze
    test_analysis.case.action: str = CaseActions.ANALYZE

    # GIVEN a database with a case with one sequenced samples for specified analysis
    link = base_store.relate_sample(test_analysis.case, test_sample, PhenotypeStatus.UNKNOWN)
    base_store.session.add(link)

    # GIVEN a cases Query
    cases: Query = base_store._get_outer_join_cases_with_analyses_query()

    # WHEN getting cases to analyse
    cases: Query = filter_cases_for_analysis(cases=cases)

    # ASSERT that cases is a query
    assert isinstance(cases, Query)

    # THEN cases should contain the test case
    assert cases


def test_filter_cases_for_analysis_when_sequenced_sample_and_no_analysis(
    base_store: Store, helpers: StoreHelpers, timestamp_now: datetime
):
    """Test that a case is returned when there are internally created cases with no action set and no prior analysis."""

    # GIVEN a sequenced sample
    test_sample: Sample = helpers.add_sample(
        base_store, is_external=False, last_sequenced_at=timestamp_now
    )

    # GIVEN a case
    test_case = helpers.add_case(base_store)

    # GIVEN a database with a case with one sequenced samples for specified analysis
    link = base_store.relate_sample(test_case, test_sample, PhenotypeStatus.UNKNOWN)
    base_store.session.add(link)

    # GIVEN a cases Query
    cases: Query = base_store._get_outer_join_cases_with_analyses_query()

    # WHEN getting cases to analyse
    cases: Query = filter_cases_for_analysis(cases=cases)

    # ASSERT that cases is a query
    assert isinstance(cases, Query)

    # THEN cases should contain the test case
    assert cases


def test_filter_cases_for_analysis_when_cases_with_no_action_and_new_sequence_data(
    base_store: Store,
    helpers: StoreHelpers,
    timestamp_yesterday: datetime,
    timestamp_now: datetime,
):
    """Test that a case is returned when cases with no action, but new sequence data."""

    # GIVEN a sequenced sample
    test_sample: Sample = helpers.add_sample(
        base_store, is_external=False, last_sequenced_at=timestamp_now
    )

    # GIVEN a completed analysis
    test_analysis: Analysis = helpers.add_analysis(base_store, workflow=Workflow.MIP_DNA)

    # Given an action set to None
    test_analysis.case.action = None

    # GIVEN a database with a case with one sequenced samples for specified analysis
    link = base_store.relate_sample(test_analysis.case, test_sample, PhenotypeStatus.UNKNOWN)
    base_store.session.add(link)

    # GIVEN an old analysis
    test_analysis.created_at = timestamp_yesterday

    # GIVEN a cases Query
    cases: Query = base_store._get_outer_join_cases_with_analyses_query()

    # WHEN getting cases to analyse
    cases: Query = filter_cases_for_analysis(cases=cases)

    # ASSERT that cases is a query
    assert isinstance(cases, Query)

    # THEN cases should contain the test case
    assert cases


def test_filter_cases_for_analysis_when_cases_with_no_action_and_old_sequence_data(
    base_store: Store, helpers: StoreHelpers, timestamp_yesterday: datetime
):
    """Test that a case is not returned when cases with no action, but old sequence data."""

    # GIVEN a sequenced sample
    test_sample: Sample = helpers.add_sample(
        base_store, is_external=True, last_sequenced_at=timestamp_yesterday
    )

    # GIVEN a completed analysis
    test_analysis: Analysis = helpers.add_analysis(base_store, workflow=Workflow.MIP_DNA)

    # Given an action set to None
    test_analysis.case.action: str | None = None

    # GIVEN a database with a case with one sequenced samples for specified analysis
    link = base_store.relate_sample(test_analysis.case, test_sample, PhenotypeStatus.UNKNOWN)
    base_store.session.add(link)

    # GIVEN a cases Query
    cases: Query = base_store._get_outer_join_cases_with_analyses_query()

    # WHEN getting cases to analyse
    cases: Query = filter_cases_for_analysis(cases=cases)

    # ASSERT that cases is a query
    assert isinstance(cases, Query)

    # THEN cases should not contain the test case
    assert not cases.all()


def test_filter_cases_with_scout_data_delivery(
    base_store: Store, helpers: StoreHelpers, timestamp_now: datetime
):
    """Test that a case is returned when Scout is specified as a data delivery option."""

    # GIVEN a sequenced sample
    test_sample: Sample = helpers.add_sample(base_store)

    # GIVEN a case with Scout as data delivery
    test_case = helpers.add_case(base_store, data_delivery=DataDelivery.FASTQ_ANALYSIS_SCOUT)

    # GIVEN a database with a case with one sequenced samples for specified analysis
    link = base_store.relate_sample(test_case, test_sample, PhenotypeStatus.UNKNOWN)
    base_store.session.add(link)

    # GIVEN a cases Query
    cases: Query = base_store._get_outer_join_cases_with_analyses_query()

    # WHEN getting cases with Scout as data delivery option
    cases: Query = filter_cases_with_scout_data_delivery(cases=cases)

    # ASSERT that cases is a query
    assert isinstance(cases, Query)

    # THEN cases should contain the test case
    assert cases


def test_filter_report_supported_data_delivery_cases(
    base_store: Store, helpers: StoreHelpers, timestamp_now: datetime
):
    """Test that a case is returned for a delivery report supported data delivery option."""

    # GIVEN a sequenced sample
    test_sample: Sample = helpers.add_sample(base_store)

    # GIVEN a case with Scout and a not supported option as data deliveries
    test_case = helpers.add_case(base_store, data_delivery=DataDelivery.FASTQ_ANALYSIS_SCOUT)
    test_invalid_case = helpers.add_case(base_store, name="test", data_delivery=DataDelivery.FASTQ)

    # GIVEN a database with the test cases
    link_1: CaseSample = base_store.relate_sample(test_case, test_sample, PhenotypeStatus.UNKNOWN)
    link_2: CaseSample = base_store.relate_sample(
        test_invalid_case, test_sample, PhenotypeStatus.UNKNOWN
    )
    base_store.session.add_all([link_1, link_2])

    # GIVEN a cases Query
    cases: Query = base_store._get_outer_join_cases_with_analyses_query()

    # WHEN retrieving the delivery report supported cases
    cases: Query = filter_report_supported_data_delivery_cases(cases=cases)

    # ASSERT that cases is a query
    assert isinstance(cases, Query)

    # THEN only the delivery report supported case should be retrieved
    assert test_case in cases
    assert test_invalid_case not in cases


def test_filter_inactive_analysis_cases(base_store: Store, helpers: StoreHelpers):
    """Test that an inactive case is returned when there is case which has no action set."""

    # GIVEN a case
    test_case = helpers.add_case(base_store)

    # GIVEN a cases Query
    cases: Query = base_store._get_query(table=Case)

    # WHEN getting completed cases
    cases: Query = filter_inactive_analysis_cases(cases=cases)

    # ASSERT that cases is a query
    assert isinstance(cases, Query)

    # THEN cases should contain the test case
    assert cases

    assert cases.all()[0].internal_id == test_case.internal_id


def test_filter_inactive_analysis_cases_when_on_hold(base_store: Store, helpers: StoreHelpers):
    """Test that an inactivated case is returned when there is case which has action set to hold."""

    # GIVEN a case
    test_case = helpers.add_case(base_store, action=CaseActions.HOLD)

    # GIVEN a cases Query
    cases: Query = base_store._get_query(table=Case)

    # WHEN getting completed cases
    cases: Query = filter_inactive_analysis_cases(cases=cases)

    # ASSERT that cases is a query
    assert isinstance(cases, Query)

    # THEN cases should contain the test case
    assert cases

    assert cases[0].internal_id == test_case.internal_id


def test_filter_inactive_analysis_cases_when_not_completed(
    base_store: Store, helpers: StoreHelpers
):
    """Test that no case is returned when there is case which action set to running."""

    # GIVEN a case
    helpers.add_case(base_store, action=CaseActions.RUNNING)

    # GIVEN a cases Query
    cases: Query = base_store._get_query(table=Case)

    # WHEN getting completed cases
    cases: Query = filter_inactive_analysis_cases(cases=cases)

    # ASSERT that cases is a query
    assert isinstance(cases, Query)

    # THEN cases should not contain the test case
    assert not cases.all()


def test_get_old_cases(base_store: Store, helpers: StoreHelpers, timestamp_in_2_weeks: datetime):
    """Test that an old case is returned when a future date is supplied."""

    # GIVEN a case
    test_case = helpers.add_case(base_store)

    # GIVEN a cases Query
    cases: Query = base_store._get_query(table=Case)

    # WHEN getting completed cases
    cases: Query = filter_older_cases_by_creation_date(
        cases=cases, creation_date=timestamp_in_2_weeks
    )

    # ASSERT that cases is a query
    assert isinstance(cases, Query)

    # THEN cases should contain the test case
    assert cases

    assert cases.all()[0].internal_id == test_case.internal_id


def test_get_old_cases_none_when_all_cases_are_too_new(
    base_store: Store, helpers: StoreHelpers, timestamp_yesterday: datetime
):
    """No cases are returned when all cases in the store are too new."""

    # GIVEN a case
    helpers.add_case(base_store)

    # GIVEN a cases Query
    cases: Query = base_store._get_query(table=Case)

    # WHEN getting completed cases
    cases: Query = filter_older_cases_by_creation_date(
        cases=cases, creation_date=timestamp_yesterday
    )

    # ASSERT that cases is a query
    assert isinstance(cases, Query)

    # THEN cases should not contain the test case
    assert not cases.all()


def test_filter_case_by_existing_entry_id(store_with_multiple_cases_and_samples: Store):
    # GIVEN a store containing a case with an entry id
    cases_query: Query = store_with_multiple_cases_and_samples._get_query(table=Case)
    entry_id: int = cases_query.first().id
    assert entry_id

    # WHEN filtering for cases with the entry_id
    cases: Query = filter_cases_by_entry_id(cases=cases_query, entry_id=entry_id)

    # THEN the case should have the entry_id
    assert cases.first().id == entry_id


def test_filter_cases_by_non_existing_entry_id(
    store_with_multiple_cases_and_samples: Store, non_existent_id: str
):
    # GIVEN a store containing cases without a specific entry id
    cases_query: Query = store_with_multiple_cases_and_samples._get_query(table=Case)
    entry_ids = [case.id for case in cases_query.all()]
    assert non_existent_id not in entry_ids

    # WHEN filtering for cases with the non existing entry id
    cases: Query = filter_cases_by_entry_id(cases=cases_query, entry_id=non_existent_id)

    # THEN the query should contain no cases
    assert cases.count() == 0


def test_filter_case_by_existing_internal_id(
    store_with_multiple_cases_and_samples: Store, case_id: str
):
    # GIVEN a store containing a case with an internal id case_id
    cases_query: Query = store_with_multiple_cases_and_samples._get_query(table=Case)
    internal_ids = [case.internal_id for case in cases_query.all()]
    assert case_id in internal_ids

    # WHEN filtering for cases with the internal id case_id
    cases: Query = filter_case_by_internal_id(cases=cases_query, internal_id=case_id)

    # THEN the query should contain only one case
    assert cases.count() == 1

    # THEN the case should have the internal id case_id
    assert cases.first().internal_id == case_id


def test_filter_cases_by_non_existing_internal_id(
    store_with_multiple_cases_and_samples: Store, non_existent_id: str
):
    # GIVEN a store containing a case with an internal id case_id
    cases_query: Query = store_with_multiple_cases_and_samples._get_query(table=Case)
    internal_ids = [case.internal_id for case in cases_query.all()]
    assert non_existent_id not in internal_ids

    # WHEN filtering for cases with the internal id case_id
    cases: Query = filter_case_by_internal_id(cases=cases_query, internal_id=non_existent_id)

    # THEN the query should contain no cases
    assert cases.count() == 0


def test_filter_case_by_empty_internal_id(store_with_multiple_cases_and_samples: Store):
    # GIVEN a store containing cases
    cases_query: Query = store_with_multiple_cases_and_samples._get_query(table=Case)

    # WHEN filtering for cases with an empty internal id
    cases: Query = filter_case_by_internal_id(cases=cases_query, internal_id="")

    # THEN the query should return no cases
    assert cases.count() == 0


def test_filter_running_cases_no_running_cases(store_with_multiple_cases_and_samples: Store):
    """Test that no cases are returned when no cases have a running action."""
    # GIVEN a store containing cases with no "running" action
    cases_query: Query = store_with_multiple_cases_and_samples._get_query(table=Case)
    cases_query = cases_query.filter(Case.action != CaseActions.RUNNING)

    # WHEN getting active cases
    active_cases: Query = filter_running_cases(cases=cases_query)

    # THEN the query should return no cases
    assert active_cases.count() == 0


def test_filter_running_cases_with_running_cases(store_with_multiple_cases_and_samples: Store):
    """Test that at least one case is returned when at least one case has a running action."""
    # GIVEN a store containing cases with at least one "running" action
    cases_query: Query = store_with_multiple_cases_and_samples._get_query(table=Case)
    actions: list[str] = [case.action for case in cases_query.all()]
    assert CaseActions.RUNNING in actions

    # WHEN getting active cases
    active_cases: Query = filter_running_cases(cases=cases_query)

    # THEN the query should return at least one case
    assert active_cases.count() >= 1


def test_filter_running_cases_only_running_cases(store_with_multiple_cases_and_samples: Store):
    """Test that all cases are returned when all cases have a running action."""
    # GIVEN a store containing only cases with "running" action
    cases_query: Query = store_with_multiple_cases_and_samples._get_query(table=Case)
    for case in cases_query.all():
        case.action = CaseActions.RUNNING

    # WHEN getting active cases
    active_cases: Query = filter_running_cases(cases=cases_query)

    # THEN the query should return the same number of cases as the original query
    assert active_cases.count() == cases_query.count()


def test_filter_cases_by_ticket_no_matching_ticket(
    store_with_multiple_cases_and_samples: Store, non_existent_id: str
):
    """Test that no cases are returned when filtering by a non-existent ticket."""
    # GIVEN a store containing cases with no matching ticket id
    cases_query: Query = store_with_multiple_cases_and_samples._get_query(table=Case)

    # WHEN filtering cases by a non-existent ticket
    filtered_cases: Query = filter_cases_by_ticket_id(cases=cases_query, ticket_id=non_existent_id)

    # THEN the query should return no cases
    assert filtered_cases.count() == 0


def test_filter_cases_by_ticket_matching_ticket(
    store_with_multiple_cases_and_samples: Store, ticket_id: str
):
    """Test that cases are returned when filtering by an existing ticket id."""
    # GIVEN a store containing cases with a matching ticket id
    cases_query: Query = store_with_multiple_cases_and_samples._get_query(table=Case)

    # WHEN filtering cases by an existing ticket id
    filtered_cases: Query = filter_cases_by_ticket_id(cases=cases_query, ticket_id=ticket_id)

    # THEN the query should return cases with the matching ticket
    assert filtered_cases.count() > 0
    for case in filtered_cases:
        assert ticket_id in case.tickets


def test_filter_cases_by_customer_entry_ids(store_with_multiple_cases_and_samples: Store):
    """Test that cases are returned when filtering by customer entry ids."""
    # GIVEN a store containing cases with customer ids
    cases_query: Query = store_with_multiple_cases_and_samples._get_query(table=Case)
    customer_ids = [case.customer_id for case in cases_query.all()]
    assert customer_ids

    # WHEN filtering cases by customer ids
    filtered_cases: Query = filter_cases_by_customer_entry_ids(
        cases=cases_query, customer_entry_ids=customer_ids
    )

    # THEN the filtered_cases should have the same count as cases_query
    assert filtered_cases.count() == cases_query.count()

    # THEN all cases in filtered_cases should have a customer_id in the customer_ids list
    for case in filtered_cases:
        assert case.customer_id in customer_ids


def test_filter_cases_by_name(store_with_multiple_cases_and_samples: Store):
    """Test that cases are returned when filtering by name."""
    # GIVEN a store containing cases with various names
    cases_query: Query = store_with_multiple_cases_and_samples._get_query(table=Case)
    test_name = cases_query.first().name

    # WHEN filtering cases by a specific name
    filtered_cases: Query = filter_cases_by_name(cases=cases_query, name=test_name)

    # THEN all cases in filtered_cases should have the specified name
    for case in filtered_cases:
        assert case.name == test_name


def test_filter_cases_by_search_pattern(store_with_multiple_cases_and_samples: Store):
    """Test that cases are returned when filtering by matching internal ids."""
    # GIVEN a store containing cases with internal ids and names
    cases_query: Query = store_with_multiple_cases_and_samples._get_query(table=Case)
    test_internal_id_pattern = cases_query.first().internal_id[:3]
    test_name_pattern = cases_query.first().name[:3]

    # WHEN filtering cases by matching internal id or name
    filtered_cases: Query = filter_cases_by_case_search(
        cases=cases_query,
        case_search=test_internal_id_pattern,
        name_search_pattern=test_name_pattern,
    )

    # THEN at least one case in filtered_cases should have an internal_id or name matching the specified patterns
    assert any(
        case.internal_id.startswith(test_internal_id_pattern)
        or case.name.startswith(test_name_pattern)
        for case in filtered_cases
    )


def test_filter_cases_not_analysed_no_cases(
    store_with_multiple_cases_and_samples: Store,
):
    """Test that no cases are returned when all cases have been analysed."""
    # GIVEN a store containing only cases that have been analysed
    cases_query: Query = store_with_multiple_cases_and_samples._get_query(table=Case)
    for case in cases_query.all():
        case.analyses.append(Analysis(completed_at=datetime.now()))

    # WHEN filtering cases for cases that have not been analysed
    filtered_cases: Query = filter_cases_not_analysed(cases=cases_query)

    # THEN the query should return no cases
    assert filtered_cases.count() == 0


def test_filter_cases_not_analysed_no_cases_in_progress(
    store_with_multiple_cases_and_samples: Store,
):
    """Test that all cases are returned when no cases are in progress."""
    # GIVEN a store containing cases that have not been analysed and no cases are in progress
    cases_query: Query = store_with_multiple_cases_and_samples._get_query(table=Case)
    for case in cases_query.all():
        case.analyses = []
        case.action = CaseActions.HOLD

    # WHEN filtering cases for cases that have not been analysed
    filtered_cases: Query = filter_cases_not_analysed(cases=cases_query)

    # THEN the query should return all cases
    assert filtered_cases.count() == cases_query.count()


def test_filter_cases_not_analysed_in_progress(
    store_with_multiple_cases_and_samples: Store,
):
    """Test that no cases in progress are returned."""
    # GIVEN a store containing cases that have not been analysed and at least one case is in progress
    cases_query: Query = store_with_multiple_cases_and_samples._get_query(table=Case)
    for case in cases_query.all():
        case.analyses = []
        case.action = CaseActions.ANALYZE

    # WHEN filtering cases for cases that have not been analysed
    filtered_cases: Query = filter_cases_not_analysed(cases=cases_query)

    # THEN the query should return no cases
    assert filtered_cases.count() == 0


def test_filter_cases_by_workflow_search_no_matching_workflow(
    store_with_multiple_cases_and_samples: Store,
):
    """Test that no cases are returned when there are no cases with matching workflow search."""
    # GIVEN a store containing cases with different workflow names
    cases_query: Query = store_with_multiple_cases_and_samples._get_query(table=Case)
    workflow_search = "non_existent_pipeline"

    # WHEN filtering cases by a non-matching workflow search
    filtered_cases: Query = filter_cases_by_workflow_search(
        cases=cases_query, workflow_search=workflow_search
    )

    # THEN the query should return no cases
    assert filtered_cases.count() == 0


def test_filter_cases_by_workflow_search_partial_match(
    store_with_multiple_cases_and_samples: Store,
):
    """Test that cases with partially matching workflow search are returned."""
    # GIVEN a store containing cases with different workflow names
    cases_query: Query = store_with_multiple_cases_and_samples._get_query(table=Case)
    workflow_search = cases_query.first().data_analysis[:3]

    # WHEN filtering cases by a partially matching workflow search
    filtered_cases: Query = filter_cases_by_workflow_search(
        cases=cases_query, workflow_search=workflow_search
    )

    # THEN the query should return the cases with partially matching workflow names
    assert filtered_cases.count() > 0
    for case in filtered_cases:
        assert workflow_search in case.data_analysis


def test_filter_cases_by_workflow_search_exact_match(
    store_with_multiple_cases_and_samples: Store,
):
    """Test that cases with exactly matching workflow search are returned."""
    # GIVEN a store containing cases with different workflow names
    cases_query: Query = store_with_multiple_cases_and_samples._get_query(table=Case)
    workflow_search = cases_query.first().data_analysis

    # WHEN filtering cases by an exactly matching workflow search
    filtered_cases: Query = filter_cases_by_workflow_search(
        cases=cases_query, workflow_search=workflow_search
    )

    # THEN the query should return the cases with exactly matching workflow names
    assert filtered_cases.count() > 0
    for case in filtered_cases:
        assert case.data_analysis == workflow_search


def test_filter_cases_by_priority_no_matching_priority(
    store_with_multiple_cases_and_samples: Store,
):
    """Test that no cases are returned when there are no cases with matching priority."""
    # GIVEN a store containing cases with different priorities
    cases_query: Query = store_with_multiple_cases_and_samples._get_query(table=Case)
    non_existent_priority = "non_existent_priority"

    # WHEN filtering cases by a non-matching priority
    filtered_cases: Query = filter_cases_by_priority(
        cases=cases_query, priority=non_existent_priority
    )

    # THEN the query should return no cases
    assert filtered_cases.count() == 0


def test_filter_cases_by_priority_matching_priority(
    store_with_multiple_cases_and_samples: Store,
):
    """Test that cases with matching priority are returned."""
    # GIVEN a store containing cases with different priorities
    cases_query: Query = store_with_multiple_cases_and_samples._get_query(table=Case)
    existing_priority = cases_query.first().priority

    # WHEN filtering cases by a matching priority
    filtered_cases: Query = filter_cases_by_priority(cases=cases_query, priority=existing_priority)

    # THEN the query should return the cases with matching priority
    assert filtered_cases.count() > 0
    for case in filtered_cases:
        assert case.priority == existing_priority


def test_filter_cases_by_priority_all_priorities(
    store_with_multiple_cases_and_samples: Store,
):
    """Test that filtering cases by all available priorities returns all cases."""
    # GIVEN a store containing cases with different priorities
    cases_query: Query = store_with_multiple_cases_and_samples._get_query(table=Case)
    all_priorities = set(case.priority for case in cases_query)

    # WHEN filtering cases by all available priorities
    filtered_cases = []
    for priority in all_priorities:
        filtered_cases.extend(filter_cases_by_priority(cases=cases_query, priority=priority).all())

    # THEN the query should return all cases
    assert len(filtered_cases) == cases_query.count()


def test_filter_newer_cases_by_order_date_no_newer_cases(
    store_with_multiple_cases_and_samples: Store,
):
    """Test that no cases are returned when there are no cases with a newer order date."""
    # GIVEN a store containing cases with different order dates
    cases_query: Query = store_with_multiple_cases_and_samples._get_query(table=Case)
    latest_order_date = max(case.ordered_at for case in cases_query)

    # WHEN filtering cases by a date that is later than the latest order date
    filtered_cases: Query = filter_newer_cases_by_order_date(
        cases=cases_query, order_date=latest_order_date
    )

    # THEN the query should return no cases
    assert filtered_cases.count() == 0


def test_filter_newer_cases_by_order_date_some_newer_cases(
    store_with_multiple_cases_and_samples: Store,
):
    """Test that cases with order dates newer than the given date are returned."""
    # GIVEN a store containing cases with different order dates
    cases_query: Query = store_with_multiple_cases_and_samples._get_query(table=Case)
    min_order_date = min(case.ordered_at for case in cases_query)
    max_order_date = max(case.ordered_at for case in cases_query)

    # Calculate an intermediate date between the minimum and maximum order dates
    some_order_date = min_order_date + (max_order_date - min_order_date) / 2

    # WHEN filtering cases by a date that is earlier than some order dates
    filtered_cases: Query = filter_newer_cases_by_order_date(
        cases=cases_query, order_date=some_order_date
    )

    # THEN the query should return the cases with order dates newer than the given date
    assert filtered_cases.count() > 0
    for case in filtered_cases:
        assert case.ordered_at > some_order_date


def test_get_older_cases_by_created_date_no_older_cases(
    store_with_multiple_cases_and_samples: Store,
):
    """Test that no cases are returned when there are no cases with an older creation date."""
    # GIVEN a store containing cases with different creation dates
    cases_query: Query = store_with_multiple_cases_and_samples._get_query(table=Case)
    oldest_created_date = min(case.created_at for case in cases_query)

    # WHEN filtering cases by a date that is later than the oldest order date
    filtered_cases: Query = filter_older_cases_by_creation_date(
        cases=cases_query, creation_date=oldest_created_date
    )

    # THEN the query should return no cases
    assert filtered_cases.count() == 0


def test_filter_runningfilter_older_cases_by_creation_date_some_newer_cases(
    store_with_multiple_cases_and_samples: Store,
):
    """Test that cases with creation dates older than the given date are returned."""
    # GIVEN a store containing cases with different creation dates
    cases_query: Query = store_with_multiple_cases_and_samples._get_query(table=Case)
    min_creation_date = min(case.created_at for case in cases_query)
    max_creation_date = max(case.created_at for case in cases_query)

    # Calculate an intermediate date between the minimum and maximum creation dates
    some_creation_date = min_creation_date + (max_creation_date - min_creation_date) / 2

    # WHEN filtering cases by a date that is earlier than some creation dates
    filtered_cases: Query = filter_older_cases_by_creation_date(
        cases=cases_query, creation_date=some_creation_date
    )

    # THEN the query should return the cases with creation dates older than the given date
    assert filtered_cases.count() > 0
    for case in filtered_cases:
        assert case.created_at < some_creation_date



import logging

import click

from cg.cli.utils import CLICK_CONTEXT_SETTINGS
from cg.constants.cli_options import DRY_RUN, FORCE
from cg.constants.nipt import Q30_THRESHOLD
from cg.exc import AnalysisUploadError
from cg.meta.upload.nipt.models import StatinaUploadFiles
from cg.meta.upload.nipt.nipt import NiptUploadAPI
from cg.models.cg_config import CGConfig

LOG = logging.getLogger(__name__)


@click.group(context_settings=CLICK_CONTEXT_SETTINGS)
def statina():
    """Upload NIPT result files to Statina"""
    pass


@statina.command("case")
@click.argument("case_id", required=True)
@DRY_RUN
@FORCE
@click.pass_obj
def batch(configs: CGConfig, case_id: str, dry_run: bool, force: bool):
    """Loading batch into the NIPT database"""

    LOG.info("*** Statina UPLOAD START ***")

    nipt_upload_api = NiptUploadAPI(configs)
    nipt_upload_api.set_dry_run(dry_run=dry_run)
    statina_files: StatinaUploadFiles = nipt_upload_api.get_statina_files(case_id=case_id)
    if dry_run:
        LOG.info(f"Found file paths for statina upload: {statina_files.json(exclude_none=True)}")
    elif force or nipt_upload_api.sequencing_run_passed_qc_value(
        case_id=case_id, q30_threshold=Q30_THRESHOLD
    ):
        nipt_upload_api.upload_to_statina_database(statina_files=statina_files)
    else:
        LOG.error(f"Uploading case failed: {case_id}")
        LOG.error(
            f"Flowcell did not pass one of the following QC parameters:\n"
            f"target_reads={nipt_upload_api.target_reads(case_id=case_id)}, Q30_threshold={Q30_THRESHOLD}"
        )
        raise AnalysisUploadError("Upload failed")



from cg.models.analysis import AnalysisModel
from cg.models.mip.mip_metrics_deliverables import MIPParsedMetrics


class MipAnalysis(AnalysisModel):
    case: str
    genome_build: str
    sample_id_metrics: list[MIPParsedMetrics]
    mip_version: str
    rank_model_version: str
    sample_ids: list[str]
    sv_rank_model_version: str



"""Utility functions for the downsampledata module."""

import logging
from pathlib import Path

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants import SequencingFileTag
from cg.utils.files import get_files_matching_pattern

LOG = logging.getLogger(__name__)


def store_downsampled_sample_bundle(
    fastq_file_output_directory: str, sample_id: str, housekeeper_api: HousekeeperAPI
) -> None:
    """
    Add a downsampled sample bundle to housekeeper and include the fastq files.
    Raises:
        FileExistsError
    """
    if not Path(fastq_file_output_directory).exists():
        raise FileExistsError(f"Cannot find: {fastq_file_output_directory}")
    create_downsampled_sample_bundle(sample_id=sample_id, housekeeper_api=housekeeper_api)
    store_downsampled_fastq_files_from_dir(
        sample_id=sample_id,
        housekeeper_api=housekeeper_api,
        fastq_file_output_directory=fastq_file_output_directory,
    )


def store_downsampled_fastq_files_from_dir(
    fastq_file_output_directory: str, sample_id: str, housekeeper_api: HousekeeperAPI
) -> None:
    """
    Add downsampled fastq files from the given directory to housekeeper.
    Raises:
        FileNotFoundError
    """
    fastq_file_paths: list[Path] = get_files_matching_pattern(
        directory=Path(fastq_file_output_directory),
        pattern=f"*{sample_id.split(sep='DS')[0]}*.{SequencingFileTag.FASTQ}.gz",
    )
    if not fastq_file_paths:
        raise FileNotFoundError(
            f"No fastq files found with pattern: *{sample_id}*.{SequencingFileTag.FASTQ}.gz"
        )
    for fastq_file_path in fastq_file_paths:
        LOG.debug(f"Found file: {fastq_file_path}")
        housekeeper_api.add_and_include_file_to_latest_version(
            bundle_name=sample_id,
            file=fastq_file_path,
            tags=[SequencingFileTag.FASTQ, sample_id],
        )


def create_downsampled_sample_bundle(housekeeper_api: HousekeeperAPI, sample_id: str) -> None:
    """Create a new bundle for the downsampled sample in housekeeper."""
    housekeeper_api.create_new_bundle_and_version(name=sample_id)



"""Module to test the validate file transfer service."""

from pathlib import Path

from cg.constants.constants import FileFormat
from cg.services.validate_file_transfer_service.validate_file_transfer_service import (
    ValidateFileTransferService,
)


def test_get_manifest_content(
    validate_file_transfer_service: ValidateFileTransferService,
    manifest_file: Path,
):
    """Test the get manifest file content method."""
    # GIVEN a manifest file

    # WHEN getting the content of the manifest file
    manifest_content: any = validate_file_transfer_service.get_manifest_file_content(
        manifest_file=manifest_file, manifest_file_format=FileFormat.TXT
    )

    # THEN assert that the content is a list
    assert isinstance(manifest_content, list)


def test_get_file_names_from_content(
    validate_file_transfer_service: ValidateFileTransferService,
    manifest_file: Path,
    expected_file_names_in_manifest: list[str],
):
    """Test the get file names from content method."""
    # GIVEN a manifest file content
    manifest_content: dict = validate_file_transfer_service.get_manifest_file_content(
        manifest_file=manifest_file, manifest_file_format=FileFormat.TXT
    )

    # WHEN getting the file names from the content
    file_names: list[str] = validate_file_transfer_service.extract_file_names_from_manifest(
        manifest_content
    )

    # THEN assert that the file names are a list
    assert isinstance(file_names, list)
    assert len(file_names) == len(expected_file_names_in_manifest)
    for file_name in file_names:
        assert file_name in expected_file_names_in_manifest


def test_is_file_in_directory(
    validate_file_transfer_service: ValidateFileTransferService,
    transfer_source_dir: Path,
    expected_file_names_in_manifest: list[str],
):
    """Test the is file in directory method."""
    # GIVEN a source directory and a list of file names
    for file_name in expected_file_names_in_manifest:
        # WHEN checking if the file is in the directory
        is_file_in_directory: bool = validate_file_transfer_service.is_file_in_directory_tree(
            file_name=file_name, source_dir=transfer_source_dir
        )

        # THEN assert that the file is in the directory
        assert is_file_in_directory


def test_validate_by_manifest_file(
    validate_file_transfer_service: ValidateFileTransferService,
    manifest_file: Path,
    transfer_source_dir: Path,
):
    """Test the validate by manifest file method."""
    # GIVEN a manifest file and a source directory

    # WHEN validating the files in the manifest file
    is_valid: bool = validate_file_transfer_service.is_transfer_completed(
        manifest_file=manifest_file,
        source_dir=transfer_source_dir,
        manifest_file_format=FileFormat.TXT,
    )

    # THEN assert that the files in the manifest are in the directory
    assert is_valid


def test_validate_by_manifest_file_fail(
    validate_file_transfer_service: ValidateFileTransferService,
    manifest_file_fail: Path,
    transfer_source_dir: Path,
):
    """Test the validate by manifest file method."""
    # GIVEN a manifest file and a source directory

    # WHEN validating the files in the manifest file
    is_valid: bool = validate_file_transfer_service.is_transfer_completed(
        manifest_file=manifest_file_fail,
        source_dir=transfer_source_dir,
        manifest_file_format=FileFormat.TXT,
    )

    # THEN assert that the files in the manifest are not in the directory
    assert not is_valid


def test_get_manifest_file_paths(
    expected_file_names_in_manifest: list[str], transfer_source_dir: Path
):
    """Test the get manifest file paths method."""
    # GIVEN a source directory
    validate_file_transfer_service = ValidateFileTransferService()

    # WHEN getting the using a pattern
    manifest_file_paths: list[Path] = validate_file_transfer_service.get_manifest_file_paths(
        source_dir=transfer_source_dir, pattern="file"
    )

    # THEN assert that the manifest file paths are a list
    assert len(manifest_file_paths) == len(expected_file_names_in_manifest)
    for manifest_file_path in manifest_file_paths:
        assert manifest_file_path.name in expected_file_names_in_manifest



"""Constants related to PacBio sequencing."""


class PacBioDirsAndFiles:
    ADAPTER_REPORT: str = "adapter.report.json"
    BASECALLING_REPORT: str = "ccs.report.json"
    CONTROL_REPORT: str = "control.report.json"
    LOADING_REPORT: str = "loading.report.json"
    RAW_DATA_REPORT: str = "raw_data.report.json"
    SMRTLINK_DATASETS_REPORT: str = "smrtlink-datasets.json"
    CCS_REPORT_SUFFIX: str = "ccs_report.json"


class CCSAttributeIDs:
    NUMBER_OF_READS: str = "ccs2.number_of_ccs_reads"
    TOTAL_NUMBER_OF_BASES: str = "ccs2.total_number_of_ccs_bases"
    MEAN_READ_LENGTH: str = "ccs2.mean_ccs_readlength"
    MEDIAN_READ_LENGTH: str = "ccs2.median_ccs_readlength"
    READ_LENGTH_N50: str = "ccs2.ccs_readlength_n50"
    MEDIAN_ACCURACY: str = "ccs2.median_accuracy"
    PERCENT_Q30: str = "ccs2.percent_ccs_bases_q30"


class ControlAttributeIDs:
    NUMBER_OF_READS: str = "control.reads_n"
    MEAN_READ_LENGTH: str = "control.readlength_mean"
    PERCENT_MEAN_READ_CONCORDANCE: str = "control.concordance_mean"
    PERCENT_MODE_READ_CONCORDANCE: str = "control.concordance_mode"


class LoadingAttributesIDs:
    PRODUCTIVE_ZMWS: str = "loading_xml_report.productive_zmws"
    P_0: str = "loading_xml_report.productivity_0_n"
    P_1: str = "loading_xml_report.productivity_1_n"
    P_2: str = "loading_xml_report.productivity_2_n"


class PolymeraseDataAttributeIDs:
    MEAN_READ_LENGTH: str = "raw_data_report.read_length"
    READ_LENGTH_N50: str = "raw_data_report.read_n50"
    MEAN_LONGEST_SUBREAD_LENGTH: str = "raw_data_report.insert_length"
    LONGEST_SUBREAD_LENGTH_N50: str = "raw_data_report.insert_n50"


class SmrtLinkDatabasesIDs:
    BIO_SAMPLE_NAME: str = "bioSampleName"
    CELL_ID: str = "cellId"
    CELL_INDEX: str = "cellIndex"
    MOVIE_NAME: str = "metadataContextId"
    PATH: str = "path"
    WELL_NAME: str = "wellName"
    WELL_SAMPLE_NAME: str = "wellSampleName"



from functools import wraps
from pathlib import Path
from typing import Any, Type

from cg.constants.constants import FileFormat
from cg.constants.pacbio import PacBioDirsAndFiles
from cg.exc import PacBioMetricsParsingError
from cg.io.controller import ReadFile
from cg.services.pacbio.metrics.models import (
    BaseMetrics,
    ControlMetrics,
    HiFiMetrics,
    PolymeraseMetrics,
    ProductivityMetrics,
    SmrtlinkDatasetsMetrics,
)
from cg.utils.files import get_file_in_directory


def handle_pac_bio_parsing_errors(func):
    """Decorator to catch any metrics parsing error to raise a PacBioMetricsParsingError instead."""

    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except FileNotFoundError as error:
            raise PacBioMetricsParsingError(f"Could not find the metrics file: {error}")
        except Exception as error:
            raise PacBioMetricsParsingError(f"An error occurred while parsing the metrics: {error}")

    return wrapper


@handle_pac_bio_parsing_errors
def parse_report_to_model(report_file: Path, data_model: Type[BaseMetrics]) -> BaseMetrics:
    """Parse the metrics report to a data model."""
    parsed_json: dict = ReadFile.read_file[FileFormat.JSON](file_path=report_file)
    metrics: list[dict[str, Any]] = parsed_json.get("attributes")
    data: dict = {report_field["id"]: report_field["value"] for report_field in metrics}
    return data_model.model_validate(data, from_attributes=True)


@handle_pac_bio_parsing_errors
def parse_dataset_metrics(report_dir: Path) -> SmrtlinkDatasetsMetrics:
    file_name = PacBioDirsAndFiles.SMRTLINK_DATASETS_REPORT
    report: Path = get_file_in_directory(directory=report_dir, file_name=file_name)
    parsed_json: dict = ReadFile.read_file[FileFormat.JSON](report)
    data: dict = parsed_json[0]
    return SmrtlinkDatasetsMetrics.model_validate(data, from_attributes=True)


@handle_pac_bio_parsing_errors
def parse_hifi_metrics(report_dir: Path) -> HiFiMetrics:
    file_name = PacBioDirsAndFiles.BASECALLING_REPORT
    report: Path = get_file_in_directory(directory=report_dir, file_name=file_name)
    return parse_report_to_model(report_file=report, data_model=HiFiMetrics)


@handle_pac_bio_parsing_errors
def parse_control_metrics(report_dir: Path) -> ControlMetrics:
    file_name = PacBioDirsAndFiles.CONTROL_REPORT
    report: Path = get_file_in_directory(directory=report_dir, file_name=file_name)
    return parse_report_to_model(report_file=report, data_model=ControlMetrics)


@handle_pac_bio_parsing_errors
def parse_productivity_metrics(report_dir: Path) -> ProductivityMetrics:
    file_name = PacBioDirsAndFiles.LOADING_REPORT
    report: Path = get_file_in_directory(directory=report_dir, file_name=file_name)
    return parse_report_to_model(report_file=report, data_model=ProductivityMetrics)


@handle_pac_bio_parsing_errors
def parse_polymerase_metrics(report_dir: Path) -> PolymeraseMetrics:
    file_name = PacBioDirsAndFiles.RAW_DATA_REPORT
    report: Path = get_file_in_directory(directory=report_dir, file_name=file_name)
    return parse_report_to_model(report_file=report, data_model=PolymeraseMetrics)



from cg.store.models import Case, CaseSample, Sample, IlluminaFlowCell, IlluminaSequencingRun
from cg.store.store import Store


def test_store_api_delete_relationships_between_sample_and_cases(
    sample_id_in_single_case: str,
    sample_id_in_multiple_cases: str,
    store_with_multiple_cases_and_samples: Store,
):
    """Test function to delete association between a sample and a case in store."""

    # GIVEN a store containing multiple samples associated with different cases
    sample_in_single_case: Sample = store_with_multiple_cases_and_samples.get_sample_by_internal_id(
        internal_id=sample_id_in_single_case
    )
    sample_in_multiple_cases: Sample = (
        store_with_multiple_cases_and_samples.get_sample_by_internal_id(
            internal_id=sample_id_in_multiple_cases
        )
    )

    assert sample_in_single_case
    assert sample_in_multiple_cases

    # WHEN removing the relationships between one sample and its cases
    store_with_multiple_cases_and_samples.decouple_sample_from_cases(sample_in_single_case.id)

    # THEN it should no longer be associated with any cases, but other relationships should remain
    results: list[CaseSample] = (
        store_with_multiple_cases_and_samples._get_query(table=CaseSample)
        .filter(CaseSample.sample_id == sample_in_single_case.id)
        .all()
    )
    existing_relationships: list[CaseSample] = (
        store_with_multiple_cases_and_samples._get_query(table=CaseSample)
        .filter(CaseSample.sample_id == sample_in_multiple_cases.id)
        .all()
    )

    assert not results
    assert existing_relationships


def test_store_api_delete_all_empty_cases(
    case_id_without_samples: str,
    case_id_with_multiple_samples: str,
    store_with_multiple_cases_and_samples: Store,
):
    """Test function to delete cases that are not associated with any samples"""

    # GIVEN a database containing a case without samples and a case with samples
    case_without_samples: list[CaseSample] = (
        store_with_multiple_cases_and_samples.get_case_samples_by_case_id(
            case_internal_id=case_id_without_samples
        )
    )
    case_with_samples: list[CaseSample] = (
        store_with_multiple_cases_and_samples.get_case_samples_by_case_id(
            case_internal_id=case_id_with_multiple_samples
        )
    )

    assert not case_without_samples
    assert case_with_samples

    # WHEN removing empty cases
    store_with_multiple_cases_and_samples.delete_cases_without_samples(
        [case_id_without_samples, case_id_with_multiple_samples]
    )

    # THEN no entry should be found for the empty case, but the one with samples should remain.
    result: list[CaseSample] = store_with_multiple_cases_and_samples.get_case_samples_by_case_id(
        case_internal_id=case_id_without_samples
    )
    case_with_samples: list[CaseSample] = (
        store_with_multiple_cases_and_samples.get_case_samples_by_case_id(
            case_internal_id=case_id_with_multiple_samples
        )
    )

    assert not result
    assert case_with_samples


def test_store_api_delete_non_existing_case(
    case_id_does_not_exist: str, store_with_multiple_cases_and_samples: Store
):
    """Test that nothing happens when trying to delete a case that does not exist."""

    # GIVEN a database containing some cases but not a specific case
    case: Case = store_with_multiple_cases_and_samples.get_case_by_internal_id(
        internal_id=case_id_does_not_exist
    )
    existing_cases: list[Case] = store_with_multiple_cases_and_samples.get_cases()

    assert not case
    assert existing_cases

    # WHEN removing empty cases, specifying the non existing case
    store_with_multiple_cases_and_samples.delete_cases_without_samples(
        case_internal_ids=[case_id_does_not_exist]
    )

    # THEN no case has been deleted and nothing happens
    remaining_cases: list[Case] = store_with_multiple_cases_and_samples.get_cases()

    assert len(remaining_cases) == len(existing_cases)


def test_delete_illumina_flow_cell(
    store_with_illumina_sequencing_data: Store, novaseq_x_flow_cell_id: str
):
    # GIVEN a store containing an Illumina flow cell and sequencing data
    flow_cell: IlluminaFlowCell = (
        store_with_illumina_sequencing_data.get_illumina_flow_cell_by_internal_id(
            novaseq_x_flow_cell_id
        )
    )
    sequencing_run: IlluminaSequencingRun = (
        store_with_illumina_sequencing_data.get_illumina_sequencing_run_by_device_internal_id(
            novaseq_x_flow_cell_id
        )
    )

    assert flow_cell
    assert sequencing_run

    # WHEN deleting the flow cell
    store_with_illumina_sequencing_data.delete_illumina_flow_cell(novaseq_x_flow_cell_id)

    # THEN the flow cell should no longer be found in the store
    deleted_flow_cell: IlluminaFlowCell = (
        store_with_illumina_sequencing_data.get_illumina_flow_cell_by_internal_id(
            novaseq_x_flow_cell_id
        )
    )
    deleted_sequencing_run: IlluminaSequencingRun = (
        store_with_illumina_sequencing_data.get_illumina_sequencing_run_by_device_internal_id(
            novaseq_x_flow_cell_id
        )
    )

    assert not deleted_flow_cell
    assert not deleted_sequencing_run



from pathlib import Path

from cg.services.pacbio.metrics.models import (
    ControlMetrics,
    HiFiMetrics,
    PacBioMetrics,
    PolymeraseMetrics,
    ProductivityMetrics,
    SmrtlinkDatasetsMetrics,
)
from cg.services.pacbio.metrics.utils import (
    parse_control_metrics,
    parse_dataset_metrics,
    parse_hifi_metrics,
    parse_polymerase_metrics,
    parse_productivity_metrics,
)


class PacBioMetricsParser:
    """Class for parsing PacBio sequencing metrics."""

    @staticmethod
    def parse_metrics(smrt_cell_path: Path) -> PacBioMetrics:
        """Return all the relevant PacBio metrics parsed in a single Pydantic object."""
        report_dir = Path(smrt_cell_path, "statistics")
        hifi_metrics: HiFiMetrics = parse_hifi_metrics(report_dir)
        control_metrics: ControlMetrics = parse_control_metrics(report_dir)
        productivity_metrics: ProductivityMetrics = parse_productivity_metrics(report_dir)
        polymerase_metrics: PolymeraseMetrics = parse_polymerase_metrics(report_dir)
        dataset_metrics: SmrtlinkDatasetsMetrics = parse_dataset_metrics(report_dir)

        return PacBioMetrics(
            hifi=hifi_metrics,
            control=control_metrics,
            productivity=productivity_metrics,
            polymerase=polymerase_metrics,
            dataset_metrics=dataset_metrics,
        )



import re
from typing import Any, TypeVar

from pydantic import BaseModel, Field, field_validator, model_validator

from cg.constants.pacbio import (
    CCSAttributeIDs,
    ControlAttributeIDs,
    LoadingAttributesIDs,
    PolymeraseDataAttributeIDs,
    SmrtLinkDatabasesIDs,
)
from cg.utils.calculations import divide_by_thousand_with_one_decimal, fraction_to_percent

BaseMetrics = TypeVar("BaseMetrics", bound=BaseModel)


class HiFiMetrics(BaseModel):
    """Model for the HiFi metrics."""

    reads: int = Field(..., alias=CCSAttributeIDs.NUMBER_OF_READS)
    yield_: int = Field(..., alias=CCSAttributeIDs.TOTAL_NUMBER_OF_BASES)
    mean_read_length_kb: float = Field(..., alias=CCSAttributeIDs.MEAN_READ_LENGTH)
    median_read_length: int = Field(..., alias=CCSAttributeIDs.MEDIAN_READ_LENGTH)
    mean_length_n50: int = Field(..., alias=CCSAttributeIDs.READ_LENGTH_N50)
    median_read_quality: str = Field(..., alias=CCSAttributeIDs.MEDIAN_ACCURACY)
    percent_q30: float = Field(..., alias=CCSAttributeIDs.PERCENT_Q30)

    _validate_mean_read_length_kb = field_validator("mean_read_length_kb", mode="before")(
        divide_by_thousand_with_one_decimal
    )
    _validate_percent_q30 = field_validator("percent_q30", mode="before")(fraction_to_percent)


class ControlMetrics(BaseModel):
    """Model for the control metrics."""

    reads: int = Field(..., alias=ControlAttributeIDs.NUMBER_OF_READS)
    mean_read_length_kb: float = Field(..., alias=ControlAttributeIDs.MEAN_READ_LENGTH)
    percent_mean_concordance_reads: float = Field(
        ..., alias=ControlAttributeIDs.PERCENT_MEAN_READ_CONCORDANCE
    )
    percent_mode_concordance_reads: float = Field(
        ..., alias=ControlAttributeIDs.PERCENT_MODE_READ_CONCORDANCE
    )

    _validate_mean_read_length_kb = field_validator("mean_read_length_kb", mode="before")(
        divide_by_thousand_with_one_decimal
    )
    _validate_percent_mean_concordance_reads = field_validator(
        "percent_mean_concordance_reads", mode="before"
    )(fraction_to_percent)

    _validate_percent_mode_concordance_reads = field_validator(
        "percent_mode_concordance_reads", mode="before"
    )(fraction_to_percent)


class ProductivityMetrics(BaseModel):
    """Model for the loading metrics."""

    productive_zmws: int = Field(..., alias=LoadingAttributesIDs.PRODUCTIVE_ZMWS)
    p_0: int = Field(..., alias=LoadingAttributesIDs.P_0)
    p_1: int = Field(..., alias=LoadingAttributesIDs.P_1)
    p_2: int = Field(..., alias=LoadingAttributesIDs.P_2)
    percent_p_0: float
    percent_p_1: float
    percent_p_2: float

    @model_validator(mode="before")
    @classmethod
    def set_percentages(cls, data: Any):
        if isinstance(data, dict):
            productive_zmws = data.get(LoadingAttributesIDs.PRODUCTIVE_ZMWS)
            p_0 = data.get(LoadingAttributesIDs.P_0)
            p_1 = data.get(LoadingAttributesIDs.P_1)
            p_2 = data.get(LoadingAttributesIDs.P_2)
            data["percent_p_0"] = round((p_0 / productive_zmws) * 100, 0)
            data["percent_p_1"] = round((p_1 / productive_zmws) * 100, 0)
            data["percent_p_2"] = round((p_2 / productive_zmws) * 100, 0)
        return data


class PolymeraseMetrics(BaseModel):
    """Model for the polymerase metrics."""

    mean_read_length: float = Field(..., alias=PolymeraseDataAttributeIDs.MEAN_READ_LENGTH)
    read_length_n50: float = Field(..., alias=PolymeraseDataAttributeIDs.READ_LENGTH_N50)
    mean_longest_subread_length: float = Field(
        ..., alias=PolymeraseDataAttributeIDs.MEAN_LONGEST_SUBREAD_LENGTH
    )
    longest_subread_length_n50: float = Field(
        ..., alias=PolymeraseDataAttributeIDs.LONGEST_SUBREAD_LENGTH_N50
    )

    _validate_mean_read_length = field_validator("mean_read_length", mode="before")(
        divide_by_thousand_with_one_decimal
    )
    _validate_read_length_n50 = field_validator("read_length_n50", mode="before")(
        divide_by_thousand_with_one_decimal
    )
    _validate_mean_longest_subread_length = field_validator(
        "mean_longest_subread_length", mode="before"
    )(divide_by_thousand_with_one_decimal)
    _validate_longest_subread_length_n50 = field_validator(
        "longest_subread_length_n50", mode="before"
    )(divide_by_thousand_with_one_decimal)


class SmrtlinkDatasetsMetrics(BaseModel):
    """Model to parse metrics in the SMRTlink datasets report."""

    cell_id: str = Field(..., alias=SmrtLinkDatabasesIDs.CELL_ID)
    well: str = Field(..., alias=SmrtLinkDatabasesIDs.WELL_NAME)
    well_sample_name: str = Field(..., alias=SmrtLinkDatabasesIDs.WELL_SAMPLE_NAME)
    sample_internal_id: str = Field(..., alias=SmrtLinkDatabasesIDs.BIO_SAMPLE_NAME)
    movie_name: str = Field(..., alias=SmrtLinkDatabasesIDs.MOVIE_NAME)
    cell_index: int = Field(..., alias=SmrtLinkDatabasesIDs.CELL_INDEX)
    path: str = Field(..., alias=SmrtLinkDatabasesIDs.PATH)
    plate: int

    @model_validator(mode="before")
    @classmethod
    def extract_plate(cls, data: Any):
        if isinstance(data, dict):
            path = data.get("path")
            if path:
                pattern = r"/([12])_[ABCD]01"
                match = re.search(pattern, path)
                if match:
                    data["plate"] = match.group(1)
        return data


class PacBioMetrics(BaseModel):
    """Model that holds all relevant PacBio metrics."""

    hifi: HiFiMetrics
    control: ControlMetrics
    productivity: ProductivityMetrics
    polymerase: PolymeraseMetrics
    dataset_metrics: SmrtlinkDatasetsMetrics



from http import HTTPStatus
from flask import Blueprint, jsonify, request

from cg.server.dto.samples.collaborator_samples_request import CollaboratorSamplesRequest
from cg.server.dto.samples.samples_response import SamplesResponse
from cg.server.endpoints.utils import before_request
from cg.server.ext import sample_service


SAMPLES_BLUEPRINT = Blueprint("samples", __name__, url_prefix="/api/v1")
SAMPLES_BLUEPRINT.before_request(before_request)


@SAMPLES_BLUEPRINT.route("/samples_in_collaboration")
def get_samples_in_collaboration():
    data = CollaboratorSamplesRequest.model_validate(request.values.to_dict())
    response: SamplesResponse = sample_service.get_collaborator_samples(data)
    return jsonify(response.model_dump()), HTTPStatus.OK



"""Handler to delete data objects"""

from sqlalchemy.orm import Session

from cg.store.base import BaseHandler
from cg.store.models import Case, Sample


class DeleteDataHandler(BaseHandler):
    """Contains methods to delete business data model instances."""

    def __init__(self, session: Session):
        super().__init__(session=session)
        self.session = session

    def delete_cases_without_samples(self, case_internal_ids: list[str]) -> None:
        """Delete any cases specified in case_ids without samples."""
        for case_internal_id in case_internal_ids:
            case: Case = self.get_case_by_internal_id(internal_id=case_internal_id)
            if case and not case.links:
                self.session.delete(case)
        self.session.commit()

    def delete_illumina_flow_cell(self, internal_id: str):
        """Delete an Illumina flow cell."""
        flow_cell = self.get_illumina_flow_cell_by_internal_id(internal_id=internal_id)
        if flow_cell:
            self.session.delete(flow_cell)
            self.session.commit()
        else:
            raise ValueError(f"Illumina flow cell with internal id {internal_id} not found.")

    def decouple_sample_from_cases(self, sample_id: int) -> None:
        sample: Sample = self.get_sample_by_entry_id(sample_id)
        for case_sample in sample.links:
            self.session.delete(case_sample)
        self.session.commit()



"""Balsamic QC delivery report API."""

import logging

from cg.meta.report.balsamic import BalsamicReportAPI
from cg.meta.workflow.balsamic_qc import BalsamicQCAnalysisAPI
from cg.models.cg_config import CGConfig

LOG = logging.getLogger(__name__)


class BalsamicQCReportAPI(BalsamicReportAPI):
    """API to create Balsamic QC delivery reports."""

    def __init__(self, config: CGConfig, analysis_api: BalsamicQCAnalysisAPI):
        super().__init__(config=config, analysis_api=analysis_api)
        self.analysis_api: BalsamicQCAnalysisAPI = analysis_api



"""Abstract class for the sequencing times service."""

from abc import abstractmethod, ABC
from datetime import datetime

from cg.models.run_devices.illumina_run_directory_data import IlluminaRunDirectoryData


class SequencingTimesService(ABC):
    """Abstract class for the sequencing times service."""

    @abstractmethod
    def get_start_time(self, run_directory_data: IlluminaRunDirectoryData) -> datetime:
        """Get the start time of the sequencing run."""
        pass

    @abstractmethod
    def get_end_time(self, run_directory_data: IlluminaRunDirectoryData) -> datetime:
        """Get the end time of the sequencing run."""
        pass
